"""Lists of available environments."""
CLASSIC_CONTROL = [
    "CartPole-v0",
    "CartPole-v1",
    "MountainCarContinuous-v0",
    "MountainCar-v0",
    "Pendulum-v1",
    "Acrobot-v1",
]

BOX_2D = [
    "LunarLander-v2",
    "LunarLanderContinuous-v2",
    "BipedalWalker-v3",
    "BipedalWalkerHardcore-v3",
    "CarRacing-v0",
    "FastLunarLander-v0",
]

try:
    import retro.data

    RETRO = retro.data.list_games()
except ImportError:
    RETRO = []

try:
    from dm_control import suite

    DM_CONTROL = list(suite.ALL_TASKS)
except (ImportError, OSError):
    DM_CONTROL = []


ATARI = [
    "ALE/Boxing-v5",
    "ALE/Boxing-ram-v5",
    "ALE/RoadRunner-v5",
    "ALE/RoadRunner-ram-v5",
    "ALE/Pitfall2-v5",
    "ALE/Pitfall2-ram-v5",
    "ALE/Amidar-v5",
    "ALE/Amidar-ram-v5",
    "ALE/Zaxxon-v5",
    "ALE/Zaxxon-ram-v5",
    "ALE/MsPacman-v5",
    "ALE/MsPacman-ram-v5",
    "ALE/Breakout-v5",
    "ALE/Breakout-ram-v5",
    "ALE/FishingDerby-v5",
    "ALE/FishingDerby-ram-v5",
    "ALE/Adventure-v5",
    "ALE/Adventure-ram-v5",
    "ALE/BankHeist-v5",
    "ALE/BankHeist-ram-v5",
    "ALE/Frostbite-v5",
    "ALE/Frostbite-ram-v5",
    "ALE/ChopperCommand-v5",
    "ALE/ChopperCommand-ram-v5",
    "ALE/Pitfall-v5",
    "ALE/Pitfall-ram-v5",
    "ALE/YarsRevenge-v5",
    "ALE/YarsRevenge-ram-v5",
    "ALE/KungFuMaster-v5",
    "ALE/KungFuMaster-ram-v5",
    "ALE/Alien-v5",
    "ALE/Alien-ram-v5",
    "ALE/BattleZone-v5",
    "ALE/BattleZone-ram-v5",
    "ALE/Pooyan-v5",
    "ALE/Pooyan-ram-v5",
    "ALE/Phoenix-v5",
    "ALE/Phoenix-ram-v5",
    "ALE/CrazyClimber-v5",
    "ALE/CrazyClimber-ram-v5",
    "ALE/Assault-v5",
    "ALE/Assault-ram-v5",
    "ALE/Hero-v5",
    "ALE/Hero-ram-v5",
    "ALE/Venture-v5",
    "ALE/Venture-ram-v5",
    "ALE/TimePilot-v5",
    "ALE/TimePilot-ram-v5",
    "ALE/Defender-v5",
    "ALE/Defender-ram-v5",
    "ALE/Bowling-v5",
    "ALE/Bowling-ram-v5",
    "ALE/NameThisGame-v5",
    "ALE/NameThisGame-ram-v5",
    "ALE/DoubleDunk-v5",
    "ALE/DoubleDunk-ram-v5",
    "ALE/Jamesbond-v5",
    "ALE/Jamesbond-ram-v5",
    "ALE/BeamRider-v5",
    "ALE/BeamRider-ram-v5",
    "ALE/SpaceInvaders-v5",
    "ALE/SpaceInvaders-ram-v5",
    "ALE/AirRaid-v5",
    "ALE/AirRaid-ram-v5",
    "ALE/DemonAttack-v5",
    "ALE/DemonAttack-ram-v5",
    "ALE/Atlantis2-v5",
    "ALE/Atlantis2-ram-v5",
    "ALE/Kangaroo-v5",
    "ALE/Kangaroo-ram-v5",
    "ALE/Tutankham-v5",
    "ALE/Tutankham-ram-v5",
    "ALE/Freeway-v5",
    "ALE/Freeway-ram-v5",
    "ALE/Gravitar-v5",
    "ALE/Gravitar-ram-v5",
    "ALE/IceHockey-v5",
    "ALE/IceHockey-ram-v5",
    "ALE/ElevatorAction-v5",
    "ALE/ElevatorAction-ram-v5",
    "ALE/VideoPinball-v5",
    "ALE/VideoPinball-ram-v5",
    "ALE/Centipede-v5",
    "ALE/Centipede-ram-v5",
    "ALE/Berzerk-v5",
    "ALE/Berzerk-ram-v5",
    "ALE/Riverraid-v5",
    "ALE/Riverraid-ram-v5",
    "ALE/Tennis-v5",
    "ALE/Tennis-ram-v5",
    "ALE/Enduro-v5",
    "ALE/Enduro-ram-v5",
    "ALE/Qbert-v5",
    "ALE/Qbert-ram-v5",
    "ALE/UpNDown-v5",
    "ALE/UpNDown-ram-v5",
    "ALE/PrivateEye-v5",
    "ALE/PrivateEye-ram-v5",
    "ALE/Asteroids-v5",
    "ALE/Asteroids-ram-v5",
    "ALE/Carnival-v5",
    "ALE/Carnival-ram-v5",
    "ALE/StarGunner-v5",
    "ALE/StarGunner-ram-v5",
    "ALE/JourneyEscape-v5",
    "ALE/JourneyEscape-ram-v5",
    "ALE/Asterix-v5",
    "ALE/Asterix-ram-v5",
    "ALE/MontezumaRevenge-v5",
    "ALE/MontezumaRevenge-ram-v5",
    "ALE/Gopher-v5",
    "ALE/Gopher-ram-v5",
    "ALE/Skiing-v5",
    "ALE/Skiing-ram-v5",
    "ALE/Atlantis-v5",
    "ALE/Atlantis-ram-v5",
    "ALE/Solaris-v5",
    "ALE/Solaris-ram-v5",
    "ALE/Seaquest-v5",
    "ALE/Seaquest-ram-v5",
    "ALE/WizardOfWor-v5",
    "ALE/WizardOfWor-ram-v5",
    "ALE/Krull-v5",
    "ALE/Krull-ram-v5",
    "ALE/Pong-v5",
    "ALE/Pong-ram-v5",
    "ALE/Robotank-v5",
    "ALE/Robotank-ram-v5",
    "Adventure-v0",
    "AdventureDeterministic-v0",
    "AdventureNoFrameskip-v0",
    "Adventure-v4",
    "AdventureDeterministic-v4",
    "AdventureNoFrameskip-v4",
    "Adventure-ram-v0",
    "Adventure-ramDeterministic-v0",
    "Adventure-ramNoFrameskip-v0",
    "Adventure-ram-v4",
    "Adventure-ramDeterministic-v4",
    "Adventure-ramNoFrameskip-v4",
    "AirRaid-v0",
    "AirRaidDeterministic-v0",
    "AirRaidNoFrameskip-v0",
    "AirRaid-v4",
    "AirRaidDeterministic-v4",
    "AirRaidNoFrameskip-v4",
    "AirRaid-ram-v0",
    "AirRaid-ramDeterministic-v0",
    "AirRaid-ramNoFrameskip-v0",
    "AirRaid-ram-v4",
    "AirRaid-ramDeterministic-v4",
    "AirRaid-ramNoFrameskip-v4",
    "Alien-v0",
    "AlienDeterministic-v0",
    "AlienNoFrameskip-v0",
    "Alien-v4",
    "AlienDeterministic-v4",
    "AlienNoFrameskip-v4",
    "Alien-ram-v0",
    "Alien-ramDeterministic-v0",
    "Alien-ramNoFrameskip-v0",
    "Alien-ram-v4",
    "Alien-ramDeterministic-v4",
    "Alien-ramNoFrameskip-v4",
    "Amidar-v0",
    "AmidarDeterministic-v0",
    "AmidarNoFrameskip-v0",
    "Amidar-v4",
    "AmidarDeterministic-v4",
    "AmidarNoFrameskip-v4",
    "Amidar-ram-v0",
    "Amidar-ramDeterministic-v0",
    "Amidar-ramNoFrameskip-v0",
    "Amidar-ram-v4",
    "Amidar-ramDeterministic-v4",
    "Amidar-ramNoFrameskip-v4",
    "Assault-v0",
    "AssaultDeterministic-v0",
    "AssaultNoFrameskip-v0",
    "Assault-v4",
    "AssaultDeterministic-v4",
    "AssaultNoFrameskip-v4",
    "Assault-ram-v0",
    "Assault-ramDeterministic-v0",
    "Assault-ramNoFrameskip-v0",
    "Assault-ram-v4",
    "Assault-ramDeterministic-v4",
    "Assault-ramNoFrameskip-v4",
    "Asterix-v0",
    "AsterixDeterministic-v0",
    "AsterixNoFrameskip-v0",
    "Asterix-v4",
    "AsterixDeterministic-v4",
    "AsterixNoFrameskip-v4",
    "Asterix-ram-v0",
    "Asterix-ramDeterministic-v0",
    "Asterix-ramNoFrameskip-v0",
    "Asterix-ram-v4",
    "Asterix-ramDeterministic-v4",
    "Asterix-ramNoFrameskip-v4",
    "Asteroids-v0",
    "AsteroidsDeterministic-v0",
    "AsteroidsNoFrameskip-v0",
    "Asteroids-v4",
    "AsteroidsDeterministic-v4",
    "AsteroidsNoFrameskip-v4",
    "Asteroids-ram-v0",
    "Asteroids-ramDeterministic-v0",
    "Asteroids-ramNoFrameskip-v0",
    "Asteroids-ram-v4",
    "Asteroids-ramDeterministic-v4",
    "Asteroids-ramNoFrameskip-v4",
    "Atlantis-v0",
    "AtlantisDeterministic-v0",
    "AtlantisNoFrameskip-v0",
    "Atlantis-v4",
    "AtlantisDeterministic-v4",
    "AtlantisNoFrameskip-v4",
    "Atlantis-ram-v0",
    "Atlantis-ramDeterministic-v0",
    "Atlantis-ramNoFrameskip-v0",
    "Atlantis-ram-v4",
    "Atlantis-ramDeterministic-v4",
    "Atlantis-ramNoFrameskip-v4",
    "BankHeist-v0",
    "BankHeistDeterministic-v0",
    "BankHeistNoFrameskip-v0",
    "BankHeist-v4",
    "BankHeistDeterministic-v4",
    "BankHeistNoFrameskip-v4",
    "BankHeist-ram-v0",
    "BankHeist-ramDeterministic-v0",
    "BankHeist-ramNoFrameskip-v0",
    "BankHeist-ram-v4",
    "BankHeist-ramDeterministic-v4",
    "BankHeist-ramNoFrameskip-v4",
    "BattleZone-v0",
    "BattleZoneDeterministic-v0",
    "BattleZoneNoFrameskip-v0",
    "BattleZone-v4",
    "BattleZoneDeterministic-v4",
    "BattleZoneNoFrameskip-v4",
    "BattleZone-ram-v0",
    "BattleZone-ramDeterministic-v0",
    "BattleZone-ramNoFrameskip-v0",
    "BattleZone-ram-v4",
    "BattleZone-ramDeterministic-v4",
    "BattleZone-ramNoFrameskip-v4",
    "BeamRider-v0",
    "BeamRiderDeterministic-v0",
    "BeamRiderNoFrameskip-v0",
    "BeamRider-v4",
    "BeamRiderDeterministic-v4",
    "BeamRiderNoFrameskip-v4",
    "BeamRider-ram-v0",
    "BeamRider-ramDeterministic-v0",
    "BeamRider-ramNoFrameskip-v0",
    "BeamRider-ram-v4",
    "BeamRider-ramDeterministic-v4",
    "BeamRider-ramNoFrameskip-v4",
    "Berzerk-v0",
    "BerzerkDeterministic-v0",
    "BerzerkNoFrameskip-v0",
    "Berzerk-v4",
    "BerzerkDeterministic-v4",
    "BerzerkNoFrameskip-v4",
    "Berzerk-ram-v0",
    "Berzerk-ramDeterministic-v0",
    "Berzerk-ramNoFrameskip-v0",
    "Berzerk-ram-v4",
    "Berzerk-ramDeterministic-v4",
    "Berzerk-ramNoFrameskip-v4",
    "Bowling-v0",
    "BowlingDeterministic-v0",
    "BowlingNoFrameskip-v0",
    "Bowling-v4",
    "BowlingDeterministic-v4",
    "BowlingNoFrameskip-v4",
    "Bowling-ram-v0",
    "Bowling-ramDeterministic-v0",
    "Bowling-ramNoFrameskip-v0",
    "Bowling-ram-v4",
    "Bowling-ramDeterministic-v4",
    "Bowling-ramNoFrameskip-v4",
    "Boxing-v0",
    "BoxingDeterministic-v0",
    "BoxingNoFrameskip-v0",
    "Boxing-v4",
    "BoxingDeterministic-v4",
    "BoxingNoFrameskip-v4",
    "Boxing-ram-v0",
    "Boxing-ramDeterministic-v0",
    "Boxing-ramNoFrameskip-v0",
    "Boxing-ram-v4",
    "Boxing-ramDeterministic-v4",
    "Boxing-ramNoFrameskip-v4",
    "Breakout-v0",
    "BreakoutDeterministic-v0",
    "BreakoutNoFrameskip-v0",
    "Breakout-v4",
    "BreakoutDeterministic-v4",
    "BreakoutNoFrameskip-v4",
    "Breakout-ram-v0",
    "Breakout-ramDeterministic-v0",
    "Breakout-ramNoFrameskip-v0",
    "Breakout-ram-v4",
    "Breakout-ramDeterministic-v4",
    "Breakout-ramNoFrameskip-v4",
    "Carnival-v0",
    "CarnivalDeterministic-v0",
    "CarnivalNoFrameskip-v0",
    "Carnival-v4",
    "CarnivalDeterministic-v4",
    "CarnivalNoFrameskip-v4",
    "Carnival-ram-v0",
    "Carnival-ramDeterministic-v0",
    "Carnival-ramNoFrameskip-v0",
    "Carnival-ram-v4",
    "Carnival-ramDeterministic-v4",
    "Carnival-ramNoFrameskip-v4",
    "Centipede-v0",
    "CentipedeDeterministic-v0",
    "CentipedeNoFrameskip-v0",
    "Centipede-v4",
    "CentipedeDeterministic-v4",
    "CentipedeNoFrameskip-v4",
    "Centipede-ram-v0",
    "Centipede-ramDeterministic-v0",
    "Centipede-ramNoFrameskip-v0",
    "Centipede-ram-v4",
    "Centipede-ramDeterministic-v4",
    "Centipede-ramNoFrameskip-v4",
    "ChopperCommand-v0",
    "ChopperCommandDeterministic-v0",
    "ChopperCommandNoFrameskip-v0",
    "ChopperCommand-v4",
    "ChopperCommandDeterministic-v4",
    "ChopperCommandNoFrameskip-v4",
    "ChopperCommand-ram-v0",
    "ChopperCommand-ramDeterministic-v0",
    "ChopperCommand-ramNoFrameskip-v0",
    "ChopperCommand-ram-v4",
    "ChopperCommand-ramDeterministic-v4",
    "ChopperCommand-ramNoFrameskip-v4",
    "CrazyClimber-v0",
    "CrazyClimberDeterministic-v0",
    "CrazyClimberNoFrameskip-v0",
    "CrazyClimber-v4",
    "CrazyClimberDeterministic-v4",
    "CrazyClimberNoFrameskip-v4",
    "CrazyClimber-ram-v0",
    "CrazyClimber-ramDeterministic-v0",
    "CrazyClimber-ramNoFrameskip-v0",
    "CrazyClimber-ram-v4",
    "CrazyClimber-ramDeterministic-v4",
    "CrazyClimber-ramNoFrameskip-v4",
    "Defender-v0",
    "DefenderDeterministic-v0",
    "DefenderNoFrameskip-v0",
    "Defender-v4",
    "DefenderDeterministic-v4",
    "DefenderNoFrameskip-v4",
    "Defender-ram-v0",
    "Defender-ramDeterministic-v0",
    "Defender-ramNoFrameskip-v0",
    "Defender-ram-v4",
    "Defender-ramDeterministic-v4",
    "Defender-ramNoFrameskip-v4",
    "DemonAttack-v0",
    "DemonAttackDeterministic-v0",
    "DemonAttackNoFrameskip-v0",
    "DemonAttack-v4",
    "DemonAttackDeterministic-v4",
    "DemonAttackNoFrameskip-v4",
    "DemonAttack-ram-v0",
    "DemonAttack-ramDeterministic-v0",
    "DemonAttack-ramNoFrameskip-v0",
    "DemonAttack-ram-v4",
    "DemonAttack-ramDeterministic-v4",
    "DemonAttack-ramNoFrameskip-v4",
    "DoubleDunk-v0",
    "DoubleDunkDeterministic-v0",
    "DoubleDunkNoFrameskip-v0",
    "DoubleDunk-v4",
    "DoubleDunkDeterministic-v4",
    "DoubleDunkNoFrameskip-v4",
    "DoubleDunk-ram-v0",
    "DoubleDunk-ramDeterministic-v0",
    "DoubleDunk-ramNoFrameskip-v0",
    "DoubleDunk-ram-v4",
    "DoubleDunk-ramDeterministic-v4",
    "DoubleDunk-ramNoFrameskip-v4",
    "ElevatorAction-v0",
    "ElevatorActionDeterministic-v0",
    "ElevatorActionNoFrameskip-v0",
    "ElevatorAction-v4",
    "ElevatorActionDeterministic-v4",
    "ElevatorActionNoFrameskip-v4",
    "ElevatorAction-ram-v0",
    "ElevatorAction-ramDeterministic-v0",
    "ElevatorAction-ramNoFrameskip-v0",
    "ElevatorAction-ram-v4",
    "ElevatorAction-ramDeterministic-v4",
    "ElevatorAction-ramNoFrameskip-v4",
    "Enduro-v0",
    "EnduroDeterministic-v0",
    "EnduroNoFrameskip-v0",
    "Enduro-v4",
    "EnduroDeterministic-v4",
    "EnduroNoFrameskip-v4",
    "Enduro-ram-v0",
    "Enduro-ramDeterministic-v0",
    "Enduro-ramNoFrameskip-v0",
    "Enduro-ram-v4",
    "Enduro-ramDeterministic-v4",
    "Enduro-ramNoFrameskip-v4",
    "FishingDerby-v0",
    "FishingDerbyDeterministic-v0",
    "FishingDerbyNoFrameskip-v0",
    "FishingDerby-v4",
    "FishingDerbyDeterministic-v4",
    "FishingDerbyNoFrameskip-v4",
    "FishingDerby-ram-v0",
    "FishingDerby-ramDeterministic-v0",
    "FishingDerby-ramNoFrameskip-v0",
    "FishingDerby-ram-v4",
    "FishingDerby-ramDeterministic-v4",
    "FishingDerby-ramNoFrameskip-v4",
    "Freeway-v0",
    "FreewayDeterministic-v0",
    "FreewayNoFrameskip-v0",
    "Freeway-v4",
    "FreewayDeterministic-v4",
    "FreewayNoFrameskip-v4",
    "Freeway-ram-v0",
    "Freeway-ramDeterministic-v0",
    "Freeway-ramNoFrameskip-v0",
    "Freeway-ram-v4",
    "Freeway-ramDeterministic-v4",
    "Freeway-ramNoFrameskip-v4",
    "Frostbite-v0",
    "FrostbiteDeterministic-v0",
    "FrostbiteNoFrameskip-v0",
    "Frostbite-v4",
    "FrostbiteDeterministic-v4",
    "FrostbiteNoFrameskip-v4",
    "Frostbite-ram-v0",
    "Frostbite-ramDeterministic-v0",
    "Frostbite-ramNoFrameskip-v0",
    "Frostbite-ram-v4",
    "Frostbite-ramDeterministic-v4",
    "Frostbite-ramNoFrameskip-v4",
    "Gopher-v0",
    "GopherDeterministic-v0",
    "GopherNoFrameskip-v0",
    "Gopher-v4",
    "GopherDeterministic-v4",
    "GopherNoFrameskip-v4",
    "Gopher-ram-v0",
    "Gopher-ramDeterministic-v0",
    "Gopher-ramNoFrameskip-v0",
    "Gopher-ram-v4",
    "Gopher-ramDeterministic-v4",
    "Gopher-ramNoFrameskip-v4",
    "Gravitar-v0",
    "GravitarDeterministic-v0",
    "GravitarNoFrameskip-v0",
    "Gravitar-v4",
    "GravitarDeterministic-v4",
    "GravitarNoFrameskip-v4",
    "Gravitar-ram-v0",
    "Gravitar-ramDeterministic-v0",
    "Gravitar-ramNoFrameskip-v0",
    "Gravitar-ram-v4",
    "Gravitar-ramDeterministic-v4",
    "Gravitar-ramNoFrameskip-v4",
    "Hero-v0",
    "HeroDeterministic-v0",
    "HeroNoFrameskip-v0",
    "Hero-v4",
    "HeroDeterministic-v4",
    "HeroNoFrameskip-v4",
    "Hero-ram-v0",
    "Hero-ramDeterministic-v0",
    "Hero-ramNoFrameskip-v0",
    "Hero-ram-v4",
    "Hero-ramDeterministic-v4",
    "Hero-ramNoFrameskip-v4",
    "IceHockey-v0",
    "IceHockeyDeterministic-v0",
    "IceHockeyNoFrameskip-v0",
    "IceHockey-v4",
    "IceHockeyDeterministic-v4",
    "IceHockeyNoFrameskip-v4",
    "IceHockey-ram-v0",
    "IceHockey-ramDeterministic-v0",
    "IceHockey-ramNoFrameskip-v0",
    "IceHockey-ram-v4",
    "IceHockey-ramDeterministic-v4",
    "IceHockey-ramNoFrameskip-v4",
    "Jamesbond-v0",
    "JamesbondDeterministic-v0",
    "JamesbondNoFrameskip-v0",
    "Jamesbond-v4",
    "JamesbondDeterministic-v4",
    "JamesbondNoFrameskip-v4",
    "Jamesbond-ram-v0",
    "Jamesbond-ramDeterministic-v0",
    "Jamesbond-ramNoFrameskip-v0",
    "Jamesbond-ram-v4",
    "Jamesbond-ramDeterministic-v4",
    "Jamesbond-ramNoFrameskip-v4",
    "JourneyEscape-v0",
    "JourneyEscapeDeterministic-v0",
    "JourneyEscapeNoFrameskip-v0",
    "JourneyEscape-v4",
    "JourneyEscapeDeterministic-v4",
    "JourneyEscapeNoFrameskip-v4",
    "JourneyEscape-ram-v0",
    "JourneyEscape-ramDeterministic-v0",
    "JourneyEscape-ramNoFrameskip-v0",
    "JourneyEscape-ram-v4",
    "JourneyEscape-ramDeterministic-v4",
    "JourneyEscape-ramNoFrameskip-v4",
    "Kangaroo-v0",
    "KangarooDeterministic-v0",
    "KangarooNoFrameskip-v0",
    "Kangaroo-v4",
    "KangarooDeterministic-v4",
    "KangarooNoFrameskip-v4",
    "Kangaroo-ram-v0",
    "Kangaroo-ramDeterministic-v0",
    "Kangaroo-ramNoFrameskip-v0",
    "Kangaroo-ram-v4",
    "Kangaroo-ramDeterministic-v4",
    "Kangaroo-ramNoFrameskip-v4",
    "Krull-v0",
    "KrullDeterministic-v0",
    "KrullNoFrameskip-v0",
    "Krull-v4",
    "KrullDeterministic-v4",
    "KrullNoFrameskip-v4",
    "Krull-ram-v0",
    "Krull-ramDeterministic-v0",
    "Krull-ramNoFrameskip-v0",
    "Krull-ram-v4",
    "Krull-ramDeterministic-v4",
    "Krull-ramNoFrameskip-v4",
    "KungFuMaster-v0",
    "KungFuMasterDeterministic-v0",
    "KungFuMasterNoFrameskip-v0",
    "KungFuMaster-v4",
    "KungFuMasterDeterministic-v4",
    "KungFuMasterNoFrameskip-v4",
    "KungFuMaster-ram-v0",
    "KungFuMaster-ramDeterministic-v0",
    "KungFuMaster-ramNoFrameskip-v0",
    "KungFuMaster-ram-v4",
    "KungFuMaster-ramDeterministic-v4",
    "KungFuMaster-ramNoFrameskip-v4",
    "MontezumaRevenge-v0",
    "MontezumaRevengeDeterministic-v0",
    "MontezumaRevengeNoFrameskip-v0",
    "MontezumaRevenge-v4",
    "MontezumaRevengeDeterministic-v4",
    "MontezumaRevengeNoFrameskip-v4",
    "MontezumaRevenge-ram-v0",
    "MontezumaRevenge-ramDeterministic-v0",
    "MontezumaRevenge-ramNoFrameskip-v0",
    "MontezumaRevenge-ram-v4",
    "MontezumaRevenge-ramDeterministic-v4",
    "MontezumaRevenge-ramNoFrameskip-v4",
    "MsPacman-v0",
    "MsPacmanDeterministic-v0",
    "MsPacmanNoFrameskip-v0",
    "MsPacman-v4",
    "MsPacmanDeterministic-v4",
    "MsPacmanNoFrameskip-v4",
    "MsPacman-ram-v0",
    "MsPacman-ramDeterministic-v0",
    "MsPacman-ramNoFrameskip-v0",
    "MsPacman-ram-v4",
    "MsPacman-ramDeterministic-v4",
    "MsPacman-ramNoFrameskip-v4",
    "NameThisGame-v0",
    "NameThisGameDeterministic-v0",
    "NameThisGameNoFrameskip-v0",
    "NameThisGame-v4",
    "NameThisGameDeterministic-v4",
    "NameThisGameNoFrameskip-v4",
    "NameThisGame-ram-v0",
    "NameThisGame-ramDeterministic-v0",
    "NameThisGame-ramNoFrameskip-v0",
    "NameThisGame-ram-v4",
    "NameThisGame-ramDeterministic-v4",
    "NameThisGame-ramNoFrameskip-v4",
    "Phoenix-v0",
    "PhoenixDeterministic-v0",
    "PhoenixNoFrameskip-v0",
    "Phoenix-v4",
    "PhoenixDeterministic-v4",
    "PhoenixNoFrameskip-v4",
    "Phoenix-ram-v0",
    "Phoenix-ramDeterministic-v0",
    "Phoenix-ramNoFrameskip-v0",
    "Phoenix-ram-v4",
    "Phoenix-ramDeterministic-v4",
    "Phoenix-ramNoFrameskip-v4",
    "Pitfall-v0",
    "PitfallDeterministic-v0",
    "PitfallNoFrameskip-v0",
    "Pitfall-v4",
    "PitfallDeterministic-v4",
    "PitfallNoFrameskip-v4",
    "Pitfall-ram-v0",
    "Pitfall-ramDeterministic-v0",
    "Pitfall-ramNoFrameskip-v0",
    "Pitfall-ram-v4",
    "Pitfall-ramDeterministic-v4",
    "Pitfall-ramNoFrameskip-v4",
    "Pong-v0",
    "PongDeterministic-v0",
    "PongNoFrameskip-v0",
    "Pong-v4",
    "PongDeterministic-v4",
    "PongNoFrameskip-v4",
    "Pong-ram-v0",
    "Pong-ramDeterministic-v0",
    "Pong-ramNoFrameskip-v0",
    "Pong-ram-v4",
    "Pong-ramDeterministic-v4",
    "Pong-ramNoFrameskip-v4",
    "Pooyan-v0",
    "PooyanDeterministic-v0",
    "PooyanNoFrameskip-v0",
    "Pooyan-v4",
    "PooyanDeterministic-v4",
    "PooyanNoFrameskip-v4",
    "Pooyan-ram-v0",
    "Pooyan-ramDeterministic-v0",
    "Pooyan-ramNoFrameskip-v0",
    "Pooyan-ram-v4",
    "Pooyan-ramDeterministic-v4",
    "Pooyan-ramNoFrameskip-v4",
    "PrivateEye-v0",
    "PrivateEyeDeterministic-v0",
    "PrivateEyeNoFrameskip-v0",
    "PrivateEye-v4",
    "PrivateEyeDeterministic-v4",
    "PrivateEyeNoFrameskip-v4",
    "PrivateEye-ram-v0",
    "PrivateEye-ramDeterministic-v0",
    "PrivateEye-ramNoFrameskip-v0",
    "PrivateEye-ram-v4",
    "PrivateEye-ramDeterministic-v4",
    "PrivateEye-ramNoFrameskip-v4",
    "Qbert-v0",
    "QbertDeterministic-v0",
    "QbertNoFrameskip-v0",
    "Qbert-v4",
    "QbertDeterministic-v4",
    "QbertNoFrameskip-v4",
    "Qbert-ram-v0",
    "Qbert-ramDeterministic-v0",
    "Qbert-ramNoFrameskip-v0",
    "Qbert-ram-v4",
    "Qbert-ramDeterministic-v4",
    "Qbert-ramNoFrameskip-v4",
    "Riverraid-v0",
    "RiverraidDeterministic-v0",
    "RiverraidNoFrameskip-v0",
    "Riverraid-v4",
    "RiverraidDeterministic-v4",
    "RiverraidNoFrameskip-v4",
    "Riverraid-ram-v0",
    "Riverraid-ramDeterministic-v0",
    "Riverraid-ramNoFrameskip-v0",
    "Riverraid-ram-v4",
    "Riverraid-ramDeterministic-v4",
    "Riverraid-ramNoFrameskip-v4",
    "RoadRunner-v0",
    "RoadRunnerDeterministic-v0",
    "RoadRunnerNoFrameskip-v0",
    "RoadRunner-v4",
    "RoadRunnerDeterministic-v4",
    "RoadRunnerNoFrameskip-v4",
    "RoadRunner-ram-v0",
    "RoadRunner-ramDeterministic-v0",
    "RoadRunner-ramNoFrameskip-v0",
    "RoadRunner-ram-v4",
    "RoadRunner-ramDeterministic-v4",
    "RoadRunner-ramNoFrameskip-v4",
    "Robotank-v0",
    "RobotankDeterministic-v0",
    "RobotankNoFrameskip-v0",
    "Robotank-v4",
    "RobotankDeterministic-v4",
    "RobotankNoFrameskip-v4",
    "Robotank-ram-v0",
    "Robotank-ramDeterministic-v0",
    "Robotank-ramNoFrameskip-v0",
    "Robotank-ram-v4",
    "Robotank-ramDeterministic-v4",
    "Robotank-ramNoFrameskip-v4",
    "Seaquest-v0",
    "SeaquestDeterministic-v0",
    "SeaquestNoFrameskip-v0",
    "Seaquest-v4",
    "SeaquestDeterministic-v4",
    "SeaquestNoFrameskip-v4",
    "Seaquest-ram-v0",
    "Seaquest-ramDeterministic-v0",
    "Seaquest-ramNoFrameskip-v0",
    "Seaquest-ram-v4",
    "Seaquest-ramDeterministic-v4",
    "Seaquest-ramNoFrameskip-v4",
    "Skiing-v0",
    "SkiingDeterministic-v0",
    "SkiingNoFrameskip-v0",
    "Skiing-v4",
    "SkiingDeterministic-v4",
    "SkiingNoFrameskip-v4",
    "Skiing-ram-v0",
    "Skiing-ramDeterministic-v0",
    "Skiing-ramNoFrameskip-v0",
    "Skiing-ram-v4",
    "Skiing-ramDeterministic-v4",
    "Skiing-ramNoFrameskip-v4",
    "Solaris-v0",
    "SolarisDeterministic-v0",
    "SolarisNoFrameskip-v0",
    "Solaris-v4",
    "SolarisDeterministic-v4",
    "SolarisNoFrameskip-v4",
    "Solaris-ram-v0",
    "Solaris-ramDeterministic-v0",
    "Solaris-ramNoFrameskip-v0",
    "Solaris-ram-v4",
    "Solaris-ramDeterministic-v4",
    "Solaris-ramNoFrameskip-v4",
    "SpaceInvaders-v0",
    "SpaceInvadersDeterministic-v0",
    "SpaceInvadersNoFrameskip-v0",
    "SpaceInvaders-v4",
    "SpaceInvadersDeterministic-v4",
    "SpaceInvadersNoFrameskip-v4",
    "SpaceInvaders-ram-v0",
    "SpaceInvaders-ramDeterministic-v0",
    "SpaceInvaders-ramNoFrameskip-v0",
    "SpaceInvaders-ram-v4",
    "SpaceInvaders-ramDeterministic-v4",
    "SpaceInvaders-ramNoFrameskip-v4",
    "StarGunner-v0",
    "StarGunnerDeterministic-v0",
    "StarGunnerNoFrameskip-v0",
    "StarGunner-v4",
    "StarGunnerDeterministic-v4",
    "StarGunnerNoFrameskip-v4",
    "StarGunner-ram-v0",
    "StarGunner-ramDeterministic-v0",
    "StarGunner-ramNoFrameskip-v0",
    "StarGunner-ram-v4",
    "StarGunner-ramDeterministic-v4",
    "StarGunner-ramNoFrameskip-v4",
    "Tennis-v0",
    "TennisDeterministic-v0",
    "TennisNoFrameskip-v0",
    "Tennis-v4",
    "TennisDeterministic-v4",
    "TennisNoFrameskip-v4",
    "Tennis-ram-v0",
    "Tennis-ramDeterministic-v0",
    "Tennis-ramNoFrameskip-v0",
    "Tennis-ram-v4",
    "Tennis-ramDeterministic-v4",
    "Tennis-ramNoFrameskip-v4",
    "TimePilot-v0",
    "TimePilotDeterministic-v0",
    "TimePilotNoFrameskip-v0",
    "TimePilot-v4",
    "TimePilotDeterministic-v4",
    "TimePilotNoFrameskip-v4",
    "TimePilot-ram-v0",
    "TimePilot-ramDeterministic-v0",
    "TimePilot-ramNoFrameskip-v0",
    "TimePilot-ram-v4",
    "TimePilot-ramDeterministic-v4",
    "TimePilot-ramNoFrameskip-v4",
    "Tutankham-v0",
    "TutankhamDeterministic-v0",
    "TutankhamNoFrameskip-v0",
    "Tutankham-v4",
    "TutankhamDeterministic-v4",
    "TutankhamNoFrameskip-v4",
    "Tutankham-ram-v0",
    "Tutankham-ramDeterministic-v0",
    "Tutankham-ramNoFrameskip-v0",
    "Tutankham-ram-v4",
    "Tutankham-ramDeterministic-v4",
    "Tutankham-ramNoFrameskip-v4",
    "UpNDown-v0",
    "UpNDownDeterministic-v0",
    "UpNDownNoFrameskip-v0",
    "UpNDown-v4",
    "UpNDownDeterministic-v4",
    "UpNDownNoFrameskip-v4",
    "UpNDown-ram-v0",
    "UpNDown-ramDeterministic-v0",
    "UpNDown-ramNoFrameskip-v0",
    "UpNDown-ram-v4",
    "UpNDown-ramDeterministic-v4",
    "UpNDown-ramNoFrameskip-v4",
    "Venture-v0",
    "VentureDeterministic-v0",
    "VentureNoFrameskip-v0",
    "Venture-v4",
    "VentureDeterministic-v4",
    "VentureNoFrameskip-v4",
    "Venture-ram-v0",
    "Venture-ramDeterministic-v0",
    "Venture-ramNoFrameskip-v0",
    "Venture-ram-v4",
    "Venture-ramDeterministic-v4",
    "Venture-ramNoFrameskip-v4",
    "VideoPinball-v0",
    "VideoPinballDeterministic-v0",
    "VideoPinballNoFrameskip-v0",
    "VideoPinball-v4",
    "VideoPinballDeterministic-v4",
    "VideoPinballNoFrameskip-v4",
    "VideoPinball-ram-v0",
    "VideoPinball-ramDeterministic-v0",
    "VideoPinball-ramNoFrameskip-v0",
    "VideoPinball-ram-v4",
    "VideoPinball-ramDeterministic-v4",
    "VideoPinball-ramNoFrameskip-v4",
    "WizardOfWor-v0",
    "WizardOfWorDeterministic-v0",
    "WizardOfWorNoFrameskip-v0",
    "WizardOfWor-v4",
    "WizardOfWorDeterministic-v4",
    "WizardOfWorNoFrameskip-v4",
    "WizardOfWor-ram-v0",
    "WizardOfWor-ramDeterministic-v0",
    "WizardOfWor-ramNoFrameskip-v0",
    "WizardOfWor-ram-v4",
    "WizardOfWor-ramDeterministic-v4",
    "WizardOfWor-ramNoFrameskip-v4",
    "YarsRevenge-v0",
    "YarsRevengeDeterministic-v0",
    "YarsRevengeNoFrameskip-v0",
    "YarsRevenge-v4",
    "YarsRevengeDeterministic-v4",
    "YarsRevengeNoFrameskip-v4",
    "YarsRevenge-ram-v0",
    "YarsRevenge-ramDeterministic-v0",
    "YarsRevenge-ramNoFrameskip-v0",
    "YarsRevenge-ram-v4",
    "YarsRevenge-ramDeterministic-v4",
    "YarsRevenge-ramNoFrameskip-v4",
    "Zaxxon-v0",
    "ZaxxonDeterministic-v0",
    "ZaxxonNoFrameskip-v0",
    "Zaxxon-v4",
    "ZaxxonDeterministic-v4",
    "ZaxxonNoFrameskip-v4",
    "Zaxxon-ram-v0",
    "Zaxxon-ramDeterministic-v0",
    "Zaxxon-ramNoFrameskip-v0",
    "Zaxxon-ram-v4",
    "Zaxxon-ramDeterministic-v4",
    "Zaxxon-ramNoFrameskip-v4",
]



"""Implement the ``plangym`` API for retro environments."""
from typing import Any, Dict, Iterable, Optional

import gym
from gym import spaces
import numpy

from plangym.core import wrap_callable
from plangym.videogames.env import VideogameEnv


class ActionDiscretizer(gym.ActionWrapper):
    """Wrap a gym-retro environment and make it use discrete actions for the Sonic game."""

    def __init__(self, env):
        """Initialize a :class`ActionDiscretizer`."""
        super(ActionDiscretizer, self).__init__(env)
        buttons = ["B", "A", "MODE", "START", "UP", "DOWN", "LEFT", "RIGHT", "C", "Y", "X", "Z"]
        actions = [
            ["LEFT"],
            ["RIGHT"],
            ["LEFT", "DOWN"],
            ["RIGHT", "DOWN"],
            ["DOWN"],
            ["DOWN", "B"],
            ["B"],
        ]
        self._actions = []
        for action in actions:
            arr = numpy.array([False] * 12)
            for button in action:
                arr[buttons.index(button)] = True
            self._actions.append(arr)
        self.action_space = spaces.Discrete(len(self._actions))

    def action(self, a) -> int:  # pylint: disable=W0221
        """Return the corresponding action in the emulator's format."""
        return self._actions[a].copy()


class RetroEnv(VideogameEnv):
    """Environment for playing ``gym-retro`` games."""

    AVAILABLE_OBS_TYPES = {"coords", "rgb", "grayscale", "ram"}
    SINGLETON = True

    def __init__(
        self,
        name: str,
        frameskip: int = 5,
        episodic_life: bool = False,
        autoreset: bool = True,
        delay_setup: bool = False,
        remove_time_limit: bool = True,
        obs_type: str = "rgb",  # ram | rgb | grayscale
        render_mode: Optional[str] = None,  # None | human | rgb_array
        wrappers: Iterable[wrap_callable] = None,
        **kwargs,
    ):
        """
        Initialize a :class:`RetroEnv`.

        Args:
            name: Name of the environment. Follows standard gym syntax conventions.
            frameskip: Number of times an action will be applied for each step \
                in dt.
            episodic_life: Return ``end = True`` when losing a life.
            autoreset: Restart environment when reaching a terminal state.
            delay_setup: If ``True`` do not initialize the ``gym.Environment`` \
                     and wait for ``setup`` to be called later.
            remove_time_limit: If True, remove the time limit from the environment.
            obs_type: One of {"rgb", "ram", "grayscale"}.
            render_mode: One of {None, "human", "rgb_aray"}.
            wrappers: Wrappers that will be applied to the underlying OpenAI env. \
                     Every element of the iterable can be either a :class:`gym.Wrapper` \
                     or a tuple containing ``(gym.Wrapper, kwargs)``.
        """
        super(RetroEnv, self).__init__(
            name=name,
            frameskip=frameskip,
            episodic_life=episodic_life,
            autoreset=autoreset,
            delay_setup=delay_setup,
            remove_time_limit=remove_time_limit,
            obs_type=obs_type,  # ram | rgb | grayscale
            render_mode=render_mode,  # None | human | rgb_array
            wrappers=wrappers,
            **kwargs,
        )

    def __getattr__(self, item):
        """Forward getattr to self.gym_env."""
        return getattr(self.gym_env, item)

    @staticmethod
    def get_win_condition(info: Dict[str, Any]) -> bool:  # pragma: no cover
        """Get win condition for games that have the end of the screen available."""
        end_screen = info.get("screen_x", 0) >= info.get("screen_x_end", 1e6)
        terminal = info.get("x", 0) >= info.get("screen_x_end", 1e6) or end_screen
        return terminal

    def get_ram(self) -> numpy.ndarray:
        """Return the ram of the emulator as a numpy array."""
        return self.get_state()  # .copy()

    def clone(self, **kwargs) -> "RetroEnv":
        """Return a copy of the environment with its initialization delayed."""
        default_kwargs = dict(
            name=self.name,
            frameskip=self.frameskip,
            wrappers=self._wrappers,
            episodic_life=self.episodic_life,
            autoreset=self.autoreset,
            delay_setup=self.delay_setup,
            obs_type=self.obs_type,
        )
        default_kwargs.update(kwargs)
        return super(RetroEnv, self).clone(**default_kwargs)

    def init_gym_env(self) -> gym.Env:
        """Initialize the retro environment."""
        import retro

        if self._gym_env is not None:
            self._gym_env.close()
        gym_env = retro.make(self.name, **self._gym_env_kwargs)
        return gym_env

    def get_state(self) -> numpy.ndarray:
        """Get the state of the retro environment."""
        state = self.gym_env.em.get_state()
        return numpy.frombuffer(state, dtype=numpy.uint8)

    def set_state(self, state: numpy.ndarray):
        """Set the state of the retro environment."""
        raw_state = state.tobytes()
        self.gym_env.em.set_state(raw_state)
        return state

    def close(self):
        """Close the underlying :class:`gym.Env`."""
        if hasattr(self, "_gym_env") and hasattr(self._gym_env, "close"):
            import gc

            self._gym_env.close()
            gc.collect()



"""Plangym API implementation."""
from abc import ABC
from typing import Callable, Generator, Tuple, Union

from gym.spaces import Space
import numpy

from plangym.core import PlanEnv, PlangymEnv


class VectorizedEnv(PlangymEnv, ABC):
    """
    Base class that defines the API for working with vectorized environments.

    A vectorized environment allows to step several copies of the environment in parallel
    when calling ``step_batch``.

    It creates a local copy of the environment that is the target of all the other
    methods of :class:`PlanEnv`. In practise, a :class:`VectorizedEnv`
    acts as a wrapper of an environment initialized with the provided parameters when calling
    __init__.

    """

    def __init__(
        self,
        env_class,
        name: str,
        frameskip: int = 1,
        autoreset: bool = True,
        delay_setup: bool = False,
        n_workers: int = 8,
        **kwargs,
    ):
        """
        Initialize a :class:`VectorizedEnv`.

        Args:
            env_class: Class of the environment to be wrapped.
            name: Name of the environment.
            frameskip: Number of times ``step`` will be called with the same action.
            autoreset: Ignored. Always set to True. Automatically reset the environment
                when the OpenAI environment returns ``end = True``.
            delay_setup: If ``True`` do not initialize the :class:`gym.Environment`
                and wait for ``setup`` to be called later.
            n_workers:  Number of workers that will be used to step the env.
            **kwargs: Additional keyword arguments passed to env_class.__init__.

        """
        self._n_workers = n_workers
        self._env_class = env_class
        self._env_kwargs = kwargs
        self._plangym_env: Union[PlangymEnv, PlanEnv, None] = None
        self.SINGLETON = env_class.SINGLETON if hasattr(env_class, "SINGLETON") else False
        self.STATE_IS_ARRAY = (
            env_class.STATE_IS_ARRAY if hasattr(env_class, "STATE_IS_ARRAY") else True
        )
        super(VectorizedEnv, self).__init__(
            name=name,
            frameskip=frameskip,
            autoreset=autoreset,
            delay_setup=delay_setup,
        )

    @property
    def n_workers(self) -> int:
        """Return the number of parallel processes that run ``step_batch`` in parallel."""
        return self._n_workers

    @property
    def plan_env(self) -> PlanEnv:
        """Environment that is wrapped by the current instance."""
        return self._plangym_env

    @property
    def obs_shape(self) -> Tuple[int]:
        """Tuple containing the shape of the observations returned by the Environment."""
        return self.plan_env.obs_shape

    @property
    def action_shape(self) -> Tuple[int]:
        """Tuple containing the shape of the actions applied to the Environment."""
        return self.plan_env.action_shape

    @property
    def action_space(self) -> Space:
        """Return the action_space of the environment."""
        return self.plan_env.action_space

    @property
    def observation_space(self) -> Space:
        """Return the observation_space of the environment."""
        return self.plan_env.observation_space

    @property
    def gym_env(self):
        """Return the instance of the environment that is being wrapped by plangym."""
        try:
            return self.plan_env.gym_env
        except AttributeError:
            return

    def __getattr__(self, item):
        """Forward attributes to the wrapped environment."""
        return getattr(self.plan_env, item)

    @staticmethod
    def split_similar_chunks(
        vector: Union[list, numpy.ndarray],
        n_chunks: int,
    ) -> Generator[Union[list, numpy.ndarray], None, None]:
        """
        Split an indexable object into similar chunks.

        Args:
            vector: Target indexable object to be split.
            n_chunks: Number of similar chunks.

        Returns:
            Generator that returns the chunks created after splitting the target object.

        """
        chunk_size = int(numpy.ceil(len(vector) / n_chunks))
        for i in range(0, len(vector), chunk_size):
            yield vector[i : i + chunk_size]

    @classmethod
    def batch_step_data(cls, actions, states, dt, batch_size):
        """Make batches of step data to distribute across workers."""
        no_states = states is None or states[0] is None
        states = [None] * len(actions) if no_states else states
        dt = dt if isinstance(dt, numpy.ndarray) else numpy.ones(len(states), dtype=int) * dt
        states_chunks = cls.split_similar_chunks(states, n_chunks=batch_size)
        actions_chunks = cls.split_similar_chunks(actions, n_chunks=batch_size)
        dt_chunks = cls.split_similar_chunks(dt, n_chunks=batch_size)
        return states_chunks, actions_chunks, dt_chunks

    @staticmethod
    def unpack_transitions(results: list, return_states: bool):
        """Aggregate the results of stepping across diferent workers."""
        _states, observs, rewards, terminals, infos = [], [], [], [], []
        for result in results:
            if not return_states:
                obs, rew, ends, info = result
            else:
                _sts, obs, rew, ends, info = result
                _states += _sts

            observs += obs
            rewards += rew
            terminals += ends
            infos += info
        if not return_states:
            transitions = observs, rewards, terminals, infos
        else:
            transitions = _states, observs, rewards, terminals, infos
        return transitions

    def create_env_callable(self, **kwargs) -> Callable[..., PlanEnv]:
        """Return a callable that initializes the environment that is being vectorized."""

        def create_env_callable(env_class, **env_kwargs):
            def _inner(**inner_kwargs):
                env_kwargs.update(inner_kwargs)
                return env_class(**env_kwargs)

            return _inner

        sub_env_kwargs = dict(self._env_kwargs)
        sub_env_kwargs["render_mode"] = self.render_mode if self.render_mode != "human" else None
        callable_kwargs = dict(
            env_class=self._env_class,
            name=self.name,
            frameskip=self.frameskip,
            delay_setup=self._env_class.SINGLETON,
            **sub_env_kwargs,
        )
        callable_kwargs.update(kwargs)
        return create_env_callable(**callable_kwargs)

    def setup(self) -> None:
        """Initialize the target environment with the parameters provided at __init__."""
        self._plangym_env: PlangymEnv = self.create_env_callable()()
        self._plangym_env.setup()

    def step(
        self,
        action: numpy.ndarray,
        state: numpy.ndarray = None,
        dt: int = 1,
        return_state: bool = None,
    ):
        """
        Step the environment applying a given action from an arbitrary state.

        If is not provided the signature matches the `step` method from OpenAI gym.

        Args:
            action: Array containing the action to be applied.
            state: State to be set before stepping the environment.
            dt: Consecutive number of times to apply the given action.
            return_state: Whether to return the state in the returned tuple. \
                If None, `step` will return the state if `state` was passed as a parameter.

        Returns:
            if states is `None` returns `(observs, rewards, ends, infos)` else
            `(new_states, observs, rewards, ends, infos)`.

        """
        return self.plan_env.step(action=action, state=state, dt=dt, return_state=return_state)

    def reset(self, return_state: bool = True):
        """
        Reset the environment and returns the first observation, or the first \
        (state, obs) tuple.

        Args:
            return_state: If true return a also the initial state of the env.

        Returns:
            Observation of the environment if `return_state` is False. Otherwise,
            return (state, obs) after reset.

        """
        if self.plan_env is None and self.delay_setup:
            self.setup()
        state, obs = self.plan_env.reset(return_state=True)
        self.sync_states(state)
        return (state, obs) if return_state else obs

    def get_state(self):
        """
        Recover the internal state of the simulation.

        A state completely describes the Environment at a given moment.

        Returns:
            State of the simulation.

        """
        return self.plan_env.get_state()

    def set_state(self, state):
        """
        Set the internal state of the simulation.

        Args:
            state: Target state to be set in the environment.

        """
        self.plan_env.set_state(state)
        self.sync_states(state)

    def render(self, mode="human"):
        """Render the environment using OpenGL. This wraps the OpenAI render method."""
        return self.plan_env.render(mode)

    def get_image(self) -> numpy.ndarray:
        """
        Return a numpy array containing the rendered view of the environment.

        Square matrices are interpreted as a greyscale image. Three-dimensional arrays
        are interpreted as RGB images with channels (Height, Width, RGB)
        """
        return self.plan_env.get_image()

    def step_with_dt(self, action: Union[numpy.ndarray, int, float], dt: int = 1) -> tuple:
        """
        Take ``dt`` simulation steps and make the environment evolve in multiples \
        of ``self.frameskip`` for a total of ``dt`` * ``self.frameskip`` steps.

        Args:
            action: Chosen action applied to the environment.
            dt: Consecutive number of times that the action will be applied.

        Returns:
            If state is `None` returns `(observs, reward, terminal, info)`
            else returns `(new_state, observs, reward, terminal, info)`.

        """
        return self.plan_env.step_with_dt(action=action, dt=dt)

    def sample_action(self):
        """
        Return a valid action that can be used to step the Environment.

        Implementing this method is optional, and it's only intended to make the
        testing process of the Environment easier.
        """
        return self.plan_env.sample_action()

    def step_batch(
        self,
        actions: numpy.ndarray,
        states: numpy.ndarray = None,
        dt: Union[numpy.ndarray, int] = 1,
        return_state: bool = None,
    ):
        """
        Vectorized version of the ``step`` method.

        It allows to step a vector of states and actions. The signature and
        behaviour is the same as ``step``, but taking a list of states, actions
        and dts as input.

        Args:
            actions: Iterable containing the different actions to be applied.
            states: Iterable containing the different states to be set.
            dt: int or array containing the frameskips that will be applied.
            return_state: Whether to return the state in the returned tuple. \
                If None, `step` will return the state if `state` was passed as a parameter.

        Returns:
            if states is None returns `(observs, rewards, ends, infos)` else
            `(new_states, observs, rewards, ends, infos)`.

        """
        dt_is_array = (isinstance(dt, numpy.ndarray) and dt.shape) or isinstance(dt, (list, tuple))
        dt = dt if dt_is_array else numpy.ones(len(actions), dtype=int) * dt
        return self.make_transitions(actions, states, dt, return_state=return_state)

    def clone(self, **kwargs) -> "PlanEnv":
        """Return a copy of the environment."""
        self_kwargs = dict(
            name=self.name,
            frameskip=self.frameskip,
            delay_setup=self.delay_setup,
            env_class=self._env_class,
            n_workers=self.n_workers,
            **self._env_kwargs,
        )
        self_kwargs.update(kwargs)
        env = self.__class__(**self_kwargs)
        return env

    def sync_states(self, state: None):
        """
        Synchronize the workers' states with the state of `self.gym_env`.

        Set all the states of the different workers of the internal :class:`BatchEnv`
        to the same state as the internal :class:`Environment` used to apply the
        non-vectorized steps.
        """
        raise NotImplementedError()

    def make_transitions(self, actions, states, dt, return_state: bool = None):
        """Implement the logic for stepping the environment in parallel."""
        raise NotImplementedError()



"""Plangym API implementation."""
from abc import ABC
from typing import Any, Callable, Dict, Iterable, Optional, Tuple, Union

import gym
from gym.spaces import Box, Space
from gym.wrappers.gray_scale_observation import GrayScaleObservation
import numpy

from plangym.utils import process_frame, remove_time_limit


wrap_callable = Union[Callable[[], gym.Wrapper], Tuple[Callable[..., gym.Wrapper], Dict[str, Any]]]


class PlanEnv(ABC):
    """
    Inherit from this class to adapt environments to different problems.

    Base class that establishes all needed methods and blueprints to work with
    Gym environments.
    """

    STATE_IS_ARRAY = True
    OBS_IS_ARRAY = True
    SINGLETON = False

    def __init__(
        self,
        name: str,
        frameskip: int = 1,
        autoreset: bool = True,
        delay_setup: bool = False,
        return_image: bool = False,
    ):
        """
        Initialize a :class:`Environment`.

        Args:
            name: Name of the environment.
            frameskip: Number of times ``step`` will be called with the same action.
            autoreset: Automatically reset the environment when the OpenAI environment
                returns ``end = True``.
            delay_setup: If ``True`` do not initialize the ``gym.Environment``
                and wait for ``setup`` to be called later (delayed setups are necessary
                when one requires to serialize the object environment or to have duplicated
                instances).
            return_image: If ``True`` add an "rgb" key in the `info` dictionary returned by `step`
             that contains an RGB representation of the environment state.

        """
        # Public attributes
        self._name = name
        self.frameskip = frameskip
        self.autoreset = autoreset
        self.delay_setup = delay_setup
        self._return_image = return_image
        # Attributes for tracking data during the step process
        self._n_step = 0
        self._obs_step = None
        self._reward_step = 0
        self._terminal_step = False
        self._info_step = {}
        self._action_step = None
        self._dt_step = None
        self._state_step = None
        self._return_state_step = None
        if not delay_setup:
            self.setup()

    def __del__(self):
        """Teardown the Environment when it is no longer needed."""
        return self.close()

    # Public API -----------------------------------------------------------------------------
    @property
    def name(self) -> str:
        """Return is the name of the environment."""
        return self._name

    @property
    def obs_shape(self) -> Tuple[int]:
        """Tuple containing the shape of the observations returned by the Environment."""
        raise NotImplementedError()

    @property
    def action_shape(self) -> Tuple[int]:
        """Tuple containing the shape of the actions applied to the Environment."""
        raise NotImplementedError()

    @property
    def unwrapped(self) -> "PlanEnv":
        """
        Completely unwrap this Environment.

        Returns:
            plangym.Environment: The base non-wrapped plangym.Environment instance

        """
        return self

    @property
    def return_image(self) -> bool:
        """
        Return `return_image` flag.

        If ``True`` add an "rgb" key in the `info` dictionary returned by `step` \
        that contains an RGB representation of the environment state.
        """
        return self._return_image

    def get_image(self) -> Union[None, numpy.ndarray]:
        """
        Return a numpy array containing the rendered view of the environment.

        Square matrices are interpreted as a grayscale image. Three-dimensional arrays
        are interpreted as RGB images with channels (Height, Width, RGB)
        """
        raise NotImplementedError()

    def step(
        self,
        action: Union[numpy.ndarray, int, float],
        state: numpy.ndarray = None,
        dt: int = 1,
        return_state: Optional[bool] = None,
    ) -> tuple:
        """
        Step the environment applying the supplied action.

        Optionally set the state to the supplied state before stepping it (the
        method prepares the environment in the given state, dismissing the current
        state, and applies the action afterwards).

        Take ``dt`` simulation steps and make the environment evolve in multiples \
        of ``self.frameskip`` for a total of ``dt`` * ``self.frameskip`` steps.

        In addition, the method allows the user to prepare the returned object,
        adding additional information and custom pre-processings via ``self.process_step``
        and ``self.get_step_tuple`` methods.

        Args:
            action: Chosen action applied to the environment.
            state: Set the environment to the given state before stepping it.
            dt: Consecutive number of times that the action will be applied.
            return_state: Whether to return the state in the returned tuple. \
                If None, `step` will return the state if `state` was passed as a parameter.

        Returns:
            if state is None returns ``(observs, reward, terminal, info)``
            else returns ``(new_state, observs, reward, terminal, info)``

        """
        self.begin_step(action=action, state=state, dt=dt, return_state=return_state)
        if state is not None:
            self.set_state(state)
        obs, reward, terminal, info = self.step_with_dt(action=action, dt=dt)
        obs, reward, terminal, info = self.process_step(
            obs=obs,
            reward=reward,
            terminal=terminal,
            info=info,
        )
        step_data = self.get_step_tuple(
            obs=obs,
            reward=reward,
            terminal=terminal,
            info=info,
        )
        self.run_autoreset(step_data)  # Resets at the end to preserve the environment state.
        return step_data

    def reset(
        self,
        return_state: bool = True,
    ) -> Union[numpy.ndarray, Tuple[numpy.ndarray, numpy.ndarray]]:
        """
        Restart the environment.

        Args:
            return_state: If ``True``, it will return the state of the environment.

        Returns:
            ``(state, obs)`` if ```return_state`` is ``True`` else return ``obs``.

        """
        obs = self.apply_reset()  # Returning info upon reset is not yet supported
        obs = self.process_obs(obs)
        return (self.get_state(), obs) if return_state else obs

    def step_batch(
        self,
        actions: Union[numpy.ndarray, Iterable[Union[numpy.ndarray, int]]],
        states: Union[numpy.ndarray, Iterable] = None,
        dt: Union[int, numpy.ndarray] = 1,
        return_state: bool = True,
    ) -> Tuple[Union[list, numpy.ndarray], ...]:
        """
        Allow stepping a vector of states and actions.

        Vectorized version of the `step` method. The signature and behaviour is
        the same as `step`, but taking a list of states, actions and dts as input.

        Args:
            actions: Iterable containing the different actions to be applied.
            states: Iterable containing the different states to be set.
            dt: int or array containing the consecutive that will be applied to each state.
                If array, the different values are distributed among the multiple environments
                (contrary to ``self.frameskip``, which is a common value for any instance).
            return_state: Whether to return the state in the returned tuple, depending on
                the boolean value. \
                If None, `step` will return the state if `state` was passed as a parameter.

        Returns:
            If return_state is `True`, the method returns `(new_states, observs, rewards, ends,
            infos)`. \
            If return_state is `False`, the method returns `(observs, rewards, ends, infos)`. \
            If return_state is `None`, the returned object depends on the states parameter.

        """
        dt_is_array = (isinstance(dt, numpy.ndarray) and dt.shape) or isinstance(dt, (list, tuple))
        dt = dt if dt_is_array else numpy.ones(len(actions), dtype=int) * dt
        no_states = states is None or states[0] is None
        states = [None] * len(actions) if no_states else states
        data = [
            self.step(action, state, dt=dt, return_state=return_state)
            for action, state, dt in zip(actions, states, dt)
        ]
        return tuple(list(x) for x in zip(*data))

    def clone(self, **kwargs) -> "PlanEnv":
        """Return a copy of the environment."""
        clone_kwargs = dict(
            name=self.name,
            frameskip=self.frameskip,
            autoreset=self.autoreset,
            delay_setup=self.delay_setup,
        )
        clone_kwargs.update(kwargs)
        return self.__class__(**clone_kwargs)

    def sample_action(self):  # pragma: no cover
        """
        Return a valid action that can be used to step the Environment.

        Implementing this method is optional, and it's only intended to make the
        testing process of the Environment easier.
        """
        pass

    # Internal API -----------------------------------------------------------------------------
    def step_with_dt(self, action: Union[numpy.ndarray, int, float], dt: int = 1):
        """
        Take ``dt`` simulation steps and make the environment evolve in multiples\
        of ``self.frameskip`` for a total of ``dt`` * ``self.frameskip`` steps.

        The method performs any post-processing to the data after applying the action
        to the environment via ``self.process_apply_action``.

        This method neither computes nor returns any state.

        Args:
            action: Chosen action applied to the environment.
            dt: Consecutive number of times that the action will be applied.

        Returns:
            Tuple containing ``(observs, reward, terminal, info)``.

        """
        self._n_step = 0
        for _ in range(int(dt) * self.frameskip):
            self._n_step += 1
            step_data = self.apply_action(action)  # Tuple (obs, reward, terminal, info)
            step_data = self.process_apply_action(*step_data)  # Post-processing to step_data
            self._obs_step, self._reward_step, self._terminal_step, self._info_step = step_data
            if self._terminal_step:
                break
        return step_data

    def run_autoreset(self, step_data):
        """Reset the environment automatically if needed."""
        *_, terminal, _ = step_data  # Assumes terminal, info are the last two elements
        if terminal and self.autoreset:
            self.reset(return_state=False)

    def get_step_tuple(
        self,
        obs,
        reward,
        terminal,
        info,
    ):
        """
        Prepare the tuple that step returns.

        This is a post processing state to have fine-grained control over what data \
        the current step is returning.

        By default it determines:
         - Return the state in the tuple (necessary information to save or load the game).
         - Adding the "rgb" key in the `info` dictionary containing an RGB \
         representation of the environment.

        Args:
            obs: Observation of the environment.
            reward: Reward signal.
            terminal: Boolean indicating if the environment is finished.
            info: Dictionary containing additional information about the environment.

        Returns:
            Tuple containing the environment data after calling `step`.
        """
        # Determine whether the method has to return the environment state
        default_mode = self._state_step is not None and self._return_state_step is None
        return_state = self._return_state_step or default_mode
        # Post processing
        obs = self.process_obs(
            obs=obs,
            reward=reward,
            terminal=terminal,
            info=info,
        )
        reward = self.process_reward(
            obs=obs,
            reward=reward,
            terminal=terminal,
            info=info,
        )
        terminal = self.process_terminal(
            obs=obs,
            reward=reward,
            terminal=terminal,
            info=info,
        )
        info = self.process_info(
            obs=obs,
            reward=reward,
            terminal=terminal,
            info=info,
        )
        step_data = (
            (self.get_state(), obs, reward, terminal, info)
            if return_state
            else (obs, reward, terminal, info)
        )
        return step_data

    def setup(self) -> None:
        """
        Run environment initialization.

        Including in this function all the code which makes the environment impossible
        to serialize will allow to dispatch the environment to different workers and
        initialize it once it's copied to the target process.
        """
        pass

    def begin_step(self, action=None, dt=None, state=None, return_state: bool = None):
        """Perform setup of step variables before starting `step_with_dt`."""
        self._n_step = 0
        self._obs_step = None
        self._reward_step = 0
        self._terminal_step = False
        self._info_step = {}
        self._action_step = action
        self._dt_step = dt
        self._state_step = state
        self._return_state_step = return_state

    def process_apply_action(
        self,
        obs,
        reward,
        terminal,
        info,
    ):
        """
        Perform any post-processing to the data returned by `apply_action`.

        Args:
            obs: Observation of the environment.
            reward: Reward signal.
            terminal: Boolean indicating if the environment is finished.
            info: Dictionary containing additional information about the environment.

        Returns:
            Tuple containing the processed data.
        """
        terminal = terminal or self._terminal_step
        reward = self._reward_step + reward
        info["n_step"] = int(self._n_step)
        return obs, reward, terminal, info

    def process_step(
        self,
        obs,
        reward,
        terminal,
        info,
    ):
        """
        Prepare the returned info dictionary.

        This is a post processing step to have fine-grained control over what data \
        the info dictionary contains.

        Args:
            obs: Observation of the environment.
            reward: Reward signal.
            terminal: Boolean indicating if the environment is finished.
            info: Dictionary containing additional information about the environment.

        Returns:
            Tuple containing the environment data after calling `step`.
        """
        info["n_step"] = info.get("n_step", int(self._n_step))
        info["dt"] = self._dt_step
        if self.return_image:
            info["rgb"] = self.get_image()
        return obs, reward, terminal, info

    def close(self) -> None:
        """Tear down the current environment."""
        pass

    # Developer API -----------------------------------------------------------------------------
    def process_obs(self, obs, **kwargs):
        """Perform optional computation for computing the observation returned by step."""
        return obs

    def process_reward(self, reward, **kwargs) -> float:
        """Perform optional computation for computing the reward returned by step."""
        return reward

    def process_terminal(self, terminal, **kwargs) -> bool:
        """Perform optional computation for computing the terminal flag returned by step."""
        return terminal

    def process_info(self, info, **kwargs) -> Dict[str, Any]:
        """Perform optional computation for computing the info dictionary returned by step."""
        return info

    def apply_action(self, action):
        """Evolve the environment for one time step applying the provided action."""
        raise NotImplementedError()

    def apply_reset(self, **kwargs):
        """Perform the resetting operation on the environment."""
        raise NotImplementedError()

    def get_state(self) -> Any:
        """
        Recover the internal state of the simulation.

        A state must completely describe the Environment at a given moment.
        """
        raise NotImplementedError()

    def set_state(self, state: Any) -> None:
        """
        Set the internal state of the simulation. Overwrite current state by the given argument.

        Args:
            state: Target state to be set in the environment.

        Returns:
            None

        """
        raise NotImplementedError()


class PlangymEnv(PlanEnv):
    """Base class for implementing OpenAI ``gym`` environments in ``plangym``."""

    AVAILABLE_RENDER_MODES = {"human", "rgb_array", None}
    AVAILABLE_OBS_TYPES = {"coords", "rgb", "grayscale"}
    DEFAULT_OBS_TYPE = "coords"

    def __init__(
        self,
        name: str,
        frameskip: int = 1,
        autoreset: bool = True,
        wrappers: Iterable[wrap_callable] = None,
        delay_setup: bool = False,
        remove_time_limit=True,
        render_mode: Optional[str] = None,
        episodic_life=False,
        obs_type=None,  # one of coords|rgb|grayscale|None
        return_image=False,
        **kwargs,
    ):
        """
        Initialize a :class:`PlangymEnv`.

        The user can read all private methods as instance properties.

        Args:
            name: Name of the environment. Follows standard gym syntax conventions.
            frameskip: Number of times an action will be applied for each ``dt``. Common
                argument to all environments.
            autoreset: Automatically reset the environment when the OpenAI environment
                returns ``end = True``.
            wrappers: Wrappers that will be applied to the underlying OpenAI env.
                Every element of the iterable can be either a :class:`gym.Wrapper`
                or a tuple containing ``(gym.Wrapper, kwargs)``.
            delay_setup: If ``True`` do not initialize the :class:`gym.Environment`
                and wait for ``setup`` to be called later.
            remove_time_limit: If True, remove the time limit from the environment.

        """
        self._render_mode = render_mode
        self._gym_env = None
        self._gym_env_kwargs = kwargs or {}  # Dictionary containing the gym.make arguments
        self._remove_time_limit = remove_time_limit
        self._wrappers = wrappers
        self._obs_space = None
        self._action_space = None
        if obs_type is not None:
            assert obs_type in self.AVAILABLE_OBS_TYPES, (
                f"obs_type {obs_type} is not accepted. Available "
                f"values are: {self.AVAILABLE_OBS_TYPES}"
            )
        self._obs_type = obs_type or self.DEFAULT_OBS_TYPE
        super(PlangymEnv, self).__init__(
            name=name,
            frameskip=frameskip,
            autoreset=autoreset,
            delay_setup=delay_setup,
            return_image=return_image,
        )

    def __str__(self):
        """Pretty print the environment."""
        text = (
            f"{self.__class__} {self.name} with parameters:\n"
            f"obs_type={self.obs_type}, render_mode={self.render_mode}\n"
            f"frameskip={self.frameskip}, obs_shape={self.obs_shape},\n"
            f"action_shape={self.action_shape}"
        )
        return text

    def __repr__(self):
        """Pretty print the environment."""
        return str(self)

    @property
    def gym_env(self):
        """Return the instance of the environment that is being wrapped by plangym."""
        if self._gym_env is None and not self.SINGLETON:
            self.setup()
        return self._gym_env

    @property
    def obs_shape(self) -> Tuple[int, ...]:
        """Tuple containing the shape of the *observations* returned by the Environment."""
        return self.observation_space.shape

    @property
    def obs_type(self) -> str:
        """Return the *type* of observation returned by the environment."""
        return self._obs_type

    @property
    def observation_space(self) -> Space:
        """Return the *observation_space* of the environment."""
        return self._obs_space

    @property
    def action_shape(self) -> Tuple[int, ...]:
        """Tuple containing the shape of the *actions* applied to the Environment."""
        return self.action_space.shape

    @property
    def action_space(self) -> Space:
        """Return the *action_space* of the environment."""
        return self._action_space

    @property
    def reward_range(self):
        """Return the *reward_range* of the environment."""
        if hasattr(self.gym_env, "reward_range"):
            return self.gym_env.reward_range

    @property
    def metadata(self):
        """Return the *metadata* of the environment."""
        if hasattr(self.gym_env, "metadata"):
            return self.gym_env.metadata
        return {"render_modes": [None, "human", "rgb_array"]}

    @property
    def render_mode(self) -> Union[None, str]:
        """Return how the game will be rendered. Values: None | human | rgb_array."""
        return self._render_mode

    @property
    def remove_time_limit(self) -> bool:
        """Return True if the Environment can only be stepped for a limited number of times."""
        return self._remove_time_limit

    def setup(self):
        """
        Initialize the target :class:`gym.Env` instance.

        The method calls ``self.init_gym_env`` to initialize the :class:``gym.Env`` instance.
        It removes time limits if needed and applies wrappers introduced by the user.
        """
        self._gym_env = self.init_gym_env()
        if self.remove_time_limit:
            self._gym_env = remove_time_limit(self._gym_env)
        if self._wrappers is not None:
            self.apply_wrappers(self._wrappers)
        self.init_spaces()

    def init_spaces(self):
        """Initialize the action_space and observation_space of the environment."""
        self._init_action_space()
        if self.obs_type == "rgb":
            self._init_obs_space_rgb()
        elif self.obs_type == "grayscale":
            self._init_obs_space_grayscale()
        elif self.obs_type == "coords":
            self._init_obs_space_coords()
        if self.observation_space is None:
            self._obs_space = self.gym_env.observation_space

    def _init_action_space(self):
        self._action_space = self.gym_env.action_space

    def _init_obs_space_rgb(self):
        if self.DEFAULT_OBS_TYPE == "rgb":
            self._obs_space = self.gym_env.observation_space
        else:
            img_shape = self.get_image().shape
            self._obs_space = Box(low=0, high=255, dtype=numpy.uint8, shape=img_shape)

    def _init_obs_space_grayscale(self):

        if self.DEFAULT_OBS_TYPE == "grayscale":
            self._obs_space = self.gym_env.observation_space
        elif self.DEFAULT_OBS_TYPE == "rgb":
            self._obs_space = self.gym_env.observation_space
            self._gym_env = GrayScaleObservation(self._gym_env)
            self._obs_space = self._gym_env.observation_space
        else:
            shape = self.get_image().shape
            self._obs_space = Box(low=0, high=255, dtype=numpy.uint8, shape=(shape[0], shape[1]))

    def _init_obs_space_coords(self):
        if self.DEFAULT_OBS_TYPE == "coords":
            if hasattr(self.gym_env, "observation_space"):
                self._obs_space = self.gym_env.observation_space
            else:
                raise NotImplementedError("No observation_space implemented.")
        else:
            img = self.reset(return_state=False)
            cords = self.get_coords_obs(img)
            self._obs_space = Box(
                low=-numpy.inf,
                high=numpy.inf,
                dtype=numpy.float32,
                shape=cords.shape,
            )

    def get_image(self) -> numpy.ndarray:
        """
        Return a numpy array containing the rendered view of the environment.

        Square matrices are interpreted as a greyscale image. Three-dimensional arrays
        are interpreted as RGB images with channels (Height, Width, RGB).
        """
        if hasattr(self.gym_env, "render"):
            return self.gym_env.render(mode="rgb_array")
        raise NotImplementedError()

    def apply_reset(
        self,
        return_state: bool = True,
    ) -> Union[numpy.ndarray, Tuple[numpy.ndarray, numpy.ndarray]]:
        """
        Restart the environment.

        Args:
            return_state: If ``True`` it will return the state of the environment.

        Returns:
            ``(state, obs)`` if ```return_state`` is ``True`` else return ``obs``.

        """
        if self.gym_env is None and self.delay_setup:
            self.setup()
        return self.gym_env.reset()

    def apply_action(self, action):
        """
        Evolve the environment for one time step applying the provided action.

        Accumulate rewards and calculate terminal flag after stepping the environment.
        """
        obs, reward, terminal, info = self.gym_env.step(action)
        return obs, reward, terminal, info

    def sample_action(self) -> Union[int, numpy.ndarray]:
        """Return a valid action that can be used to step the environment chosen at random."""
        if hasattr(self.action_space, "sample"):
            return self.action_space.sample()
        return self.gym_env.action_space.sample()  # pragma: no cover

    def clone(self, **kwargs) -> "PlangymEnv":
        """Return a copy of the environment."""
        env_kwargs = dict(
            wrappers=self._wrappers,
            remove_time_limit=self._remove_time_limit,
            render_mode=self.render_mode,
        )
        env_kwargs.update(kwargs)
        env_kwargs.update(self._gym_env_kwargs)
        env: PlangymEnv = super(PlangymEnv, self).clone(**env_kwargs)
        return env

    def close(self):
        """Close the underlying :class:`gym.Env`."""
        if hasattr(self, "_gym_env") and hasattr(self._gym_env, "close"):
            return self._gym_env.close()
        self._gym_env = None

    def init_gym_env(self) -> gym.Env:
        """Initialize the :class:``gym.Env`` instance that the current class is wrapping."""
        gym_env: gym.Env = gym.make(self.name, **self._gym_env_kwargs)
        gym_env.reset()
        return gym_env

    def seed(self, seed=None):
        """Seed the underlying :class:`gym.Env`."""
        if hasattr(self.gym_env, "seed"):
            return self.gym_env.seed(seed)

    def apply_wrappers(self, wrappers: Iterable[wrap_callable]):
        """Wrap the underlying OpenAI gym environment."""
        for item in wrappers:
            if isinstance(item, tuple):
                wrapper, kwargs = item
                if isinstance(kwargs, dict):
                    self.wrap(wrapper, **kwargs)
                elif isinstance(kwargs, (list, tuple)):
                    self.wrap(wrapper, *kwargs)
                else:
                    self.wrap(wrapper, kwargs)
            else:
                self.wrap(item)

    def wrap(self, wrapper: Callable, *args, **kwargs):
        """Apply a single OpenAI gym wrapper to the environment."""
        self._gym_env = wrapper(self.gym_env, *args, **kwargs)

    def render(self, mode="human"):
        """Render the environment using OpenGL. This wraps the OpenAI render method."""
        if hasattr(self.gym_env, "render"):
            return self.gym_env.render(mode=mode)
        raise NotImplementedError()

    def process_obs(self, obs, **kwargs):
        """
        Perform optional computation for computing the observation returned by step.

        This is a post processing step to have fine-grained control over the returned
        observation.
        """
        if self.obs_type == "coords":
            return self.get_coords_obs(obs, **kwargs)
        elif self.obs_type == "rgb":
            return self.get_rgb_obs(obs, **kwargs)
        elif self.obs_type == "grayscale":
            return self.get_grayscale_obs(obs, **kwargs)
        return obs

    def get_coords_obs(self, obs, **kwargs):
        """Calculate the observation returned by `step` when obs_type == "coords"."""
        if self.DEFAULT_OBS_TYPE == "coords":
            return obs
        raise NotImplementedError()

    def get_rgb_obs(self, obs, **kwargs):
        """Calculate the observation returned by `step` when obs_type == "rgb"."""
        if self.DEFAULT_OBS_TYPE == "rgb":
            return obs
        return self.get_image()

    def get_grayscale_obs(self, obs, **kwargs):
        """Calculate the observation returned by `step` when obs_type == "grayscale"."""
        obs = self.get_rgb_obs(obs, **kwargs)
        return process_frame(obs, mode="L")



import numpy
import pytest

from plangym.api_tests import batch_size, display, TestPlanEnv, TestPlangymEnv  # noqa: F401
from plangym.control.classic_control import ClassicControl
from plangym.vectorization.parallel import BatchEnv, ExternalProcess, ParallelEnv
from plangym.videogames.atari import AtariEnv


def parallel_cartpole():
    return ParallelEnv(env_class=ClassicControl, name="CartPole-v0", blocking=True, n_workers=2)


def parallel_pacman():
    return ParallelEnv(env_class=AtariEnv, name="MsPacman-ram-v0", n_workers=2)


environments = [parallel_cartpole, parallel_pacman]


@pytest.fixture(params=environments, scope="module")
def env(request) -> ClassicControl:
    return request.param()


class TestBatchEnv:
    def test_len(self, env):
        assert len(env._batch_env) == 2

    def test_getattr(self, env):
        assert isinstance(env._batch_env, BatchEnv)
        assert env._batch_env.observation_space is not None

    def test_getitem(self, env):
        assert isinstance(env._batch_env[0], ExternalProcess)

    def test_reset(self, env):
        obs = env._batch_env.reset(return_states=False)
        assert isinstance(obs, numpy.ndarray)
        indices = numpy.arange(len(env._batch_env._envs))
        state, obs = env._batch_env.reset(return_states=True, indices=indices)
        if env.STATE_IS_ARRAY:
            assert isinstance(state, numpy.ndarray)


class TestExternalProcess:
    def test_reset(self, env):
        ep = env._batch_env[0]
        obs = ep.reset(return_states=False, blocking=True)
        assert isinstance(obs, numpy.ndarray)
        state, obs = ep.reset(return_states=True, blocking=True)
        if env.STATE_IS_ARRAY:
            assert isinstance(state, numpy.ndarray)

        obs = ep.reset(return_states=False, blocking=False)()
        assert isinstance(obs, numpy.ndarray)
        state, obs = ep.reset(return_states=True, blocking=False)()
        if env.STATE_IS_ARRAY:
            assert isinstance(state, numpy.ndarray)

    def test_step(self, env):
        ep = env._batch_env[0]
        state, _ = ep.reset(return_states=True, blocking=True)
        ep.set_state(state, blocking=False)()
        action = env.sample_action()
        data = ep.step(action, dt=2, blocking=True)
        assert isinstance(data, tuple)
        state, *data = ep.step(action, state, blocking=True)
        assert len(data) > 0
        if env.STATE_IS_ARRAY:
            assert isinstance(state, numpy.ndarray)

        state, _ = ep.reset(return_states=True, blocking=False)()
        action = env.sample_action()
        data = ep.step(action, dt=2, blocking=False)()
        assert isinstance(data, tuple)
        state, *data = ep.step(action, state, blocking=False)()
        assert len(data) > 0

    def test_attributes(self, env):
        ep = env._batch_env[0]
        ep.observation_space
        ep.action_space.sample()
        ep.__getattr__("unwrapped")
        ep.unwrapped



"""plangym package installation metadata."""
from importlib.machinery import SourceFileLoader
from pathlib import Path

from setuptools import find_packages, setup


version = SourceFileLoader(
    "plangym.version",
    str(Path(__file__).parent / "plangym" / "version.py"),
).load_module()

with open(Path(__file__).with_name("README.md"), encoding="utf-8") as f:
    long_description = f.read()

extras = {
    "atari": ["ale-py>=0.7.0", "autorom[accept-rom-license]==0.4.2"],
    "retro": ["gym-retro>=0.8.0"],
    "test": ["pytest>=5.3.5", "pyvirtualdisplay>=1.3.2"],
    "ray": ["ray", "setproctitle"],
    "box2d": ["box2d-py==2.3.5", "pyglet>=1.4.0"],
}

extras["all"] = [item for group in extras.values() for item in group]

setup(
    name="plangym",
    description="Plangym is an interface to use OpenAI gym for planning problems. It extends the standard interface to allow setting and recovering the environment states.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    version=version.__version__,
    license="MIT",
    author="Guillem Duran Ballester",
    author_email="info@fragile.tech",
    url="https://github.com/FragileTech/plangym",
    keywords=["Machine learning", "artificial intelligence"],
    test_suite="tests",
    tests_require=["pytest>=5.3.5", "hypothesis>=5.6.0"],
    install_requires=[
        "numpy",
        "pillow==8.4.0",
        "opencv-python>=4.2.0.32",
    ],
    extras_require=extras,
    package_data={"": ["README.md"]},
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Software Development :: Libraries",
    ],
)



"""Functionality for instantiating the environment by passing the environment id."""
from plangym.environment_names import ATARI, BOX_2D, CLASSIC_CONTROL, DM_CONTROL, RETRO


def get_planenv_class(name, domain_name, state):
    """Return the class corresponding to the environment name."""
    # if name == "MinimalPacman-v0":
    #    return MinimalPacman
    # elif name == "MinimalPong-v0":
    #    return MinimalPong
    if name == "PlanMontezuma-v0":
        from plangym.videogames import MontezumaEnv

        return MontezumaEnv
    elif state is not None or name in set(RETRO):
        from plangym.videogames import RetroEnv

        return RetroEnv
    elif name in set(CLASSIC_CONTROL):
        from plangym.control import ClassicControl

        return ClassicControl
    elif name in set(BOX_2D):
        if name == "FastLunarLander-v0":
            from plangym.control import LunarLander

            return LunarLander
        from plangym.control import Box2DEnv

        return Box2DEnv
    elif name in ATARI:
        from plangym.videogames import AtariEnv

        return AtariEnv
    elif domain_name is not None or any(x[0] in name for x in DM_CONTROL):
        from plangym.control import DMControlEnv

        return DMControlEnv
    elif "SuperMarioBros" in name:
        from plangym.videogames import MarioEnv

        return MarioEnv
    elif "BalloonLearningEnvironment-v0":
        from plangym.control import BalloonEnv

        return BalloonEnv
    raise ValueError(f"Environment {name} is not supported.")


def get_environment_class(
    name: str = None,
    n_workers: int = None,
    ray: bool = False,
    domain_name: str = None,
    state: str = None,
):
    """Get the class and vectorized environment and PlangymEnv class from the make params."""
    env_class = get_planenv_class(name, domain_name, state)
    if ray:
        from plangym.vectorization import RayEnv

        return RayEnv, env_class
    elif n_workers is not None:
        from plangym.vectorization import ParallelEnv

        return ParallelEnv, env_class
    return None, env_class


def make(
    name: str = None,
    n_workers: int = None,
    ray: bool = False,
    domain_name: str = None,
    state: str = None,
    **kwargs,
):
    """Create the appropriate PlangymEnv from the environment name and other parameters."""
    parallel_class, env_class = get_environment_class(
        name=name,
        n_workers=n_workers,
        ray=ray,
        domain_name=domain_name,
        state=state,
    )
    kwargs["name"] = name
    if state is not None:
        kwargs["state"] = state
    if domain_name is not None:
        kwargs["domain_name"] = domain_name
    if parallel_class is not None:
        return parallel_class(env_class=env_class, n_workers=n_workers, **kwargs)
    return env_class(**kwargs)



import copy
from itertools import product
import os
from typing import Iterable
import warnings

import gym
import numpy
import pytest
from pyvirtualdisplay import Display

import plangym
from plangym.core import PlanEnv, PlangymEnv
from plangym.vectorization.env import VectorizedEnv
from plangym.videogames.env import LIFE_KEY


def generate_test_cases(
    names,
    env_class,
    n_workers_values=None,
    render_modes=None,
    obs_types=None,
    custom_tests=None,
) -> PlangymEnv:
    custom_tests = custom_tests or []
    n_workers_vals = [None] if n_workers_values is None else n_workers_values
    names = [names] if isinstance(names, str) else names
    available_render_modes = (
        [None] if os.getenv("SKIP_RENDER", False) else env_class.AVAILABLE_RENDER_MODES
    )
    available_obs_types = (
        [None] if os.getenv("SKIP_RENDER", False) else env_class.AVAILABLE_OBS_TYPES
    )
    render_modes = available_render_modes if render_modes is None else render_modes
    obs_types = available_obs_types if obs_types is None else obs_types
    for i, (n_workers, obs_type, render_mode) in enumerate(
        product(
            n_workers_vals,
            obs_types,
            render_modes,
        ),
    ):
        name = names[i % len(names)]
        if isinstance(name, tuple):
            name = "-".join(name)

        def _make_env():
            return plangym.make(
                name,
                n_workers=n_workers,
                obs_type=obs_type,
                render_mode=render_mode,
            )

        yield _make_env
    for custom_test in custom_tests:
        yield custom_test


@pytest.fixture(scope="class")
def batch_size() -> int:
    return 10


@pytest.fixture(scope="class")
def display():
    display = Display(visible=0, size=(400, 400))
    display.start()
    yield display
    display.stop()


def step_tuple_test(env, obs, reward, terminal, info, dt=None):
    obs_is_array = isinstance(obs, numpy.ndarray)
    assert obs_is_array if env.OBS_IS_ARRAY else not obs_is_array
    assert obs.shape == env.obs_shape, (obs.shape, env.obs_shape)
    assert float(reward) + 1 == float(reward) + 1
    assert isinstance(terminal, bool)
    assert isinstance(info, dict)
    assert "n_step" in info
    assert info["n_step"] <= int(dt * env.frameskip), (dt, env.frameskip, info.get("n_step", 0))
    assert "dt" in info
    if dt is not None:
        assert info["dt"] == dt
    if env.return_image:
        assert "rgb" in info
        assert isinstance(info["rgb"], numpy.ndarray)


def step_batch_tuple_test(env, batch_size, observs, rewards, terminals, infos, dt):
    assert len(rewards) == batch_size
    assert len(terminals) == batch_size
    assert len(observs) == batch_size
    assert len(infos) == batch_size

    dts = dt if isinstance(dt, (list, numpy.ndarray)) else [dt] * batch_size
    for obs, reward, terminal, info, dt in zip(list(observs), rewards, terminals, infos, dts):
        step_tuple_test(env=env, obs=obs, reward=reward, terminal=terminal, info=info, dt=dt)


class TestPlanEnv:

    CLASS_ATTRIBUTES = ("OBS_IS_ARRAY", "STATE_IS_ARRAY", "SINGLETON")
    PROPERTIES = (
        "unwrapped",
        "obs_shape",
        "action_shape",
        "name",
        "frameskip",
        "autoreset",
        "delay_setup",
        "return_image",
    )

    def test_init(self, env):
        pass

    def test_repr(self, env):
        assert str(env) == repr(env)

    # Test attributes and properties
    # ---------------------------------------------------------------------------------------------
    def test_class_attributes(self, env):
        for name in self.CLASS_ATTRIBUTES:
            assert hasattr(env.__class__, name), f"Env {env.name} does not have attribute {name}"
            isinstance(getattr(env.__class__, name), bool)

    def test_has_properties(self, env):
        for name in self.PROPERTIES:
            assert hasattr(env, name), f"Env {env.name} does not have property {name}"

    def test_name(self, env):
        assert isinstance(env.name, str)

    def test_obs_shape(self, env):
        assert hasattr(env, "obs_shape")
        assert isinstance(env.obs_shape, tuple)
        if env.obs_shape:
            for val in env.obs_shape:
                assert isinstance(val, int)
        obs = env.reset(return_state=False)
        assert obs.shape == env.obs_shape
        obs, *_ = env.step(env.sample_action())
        assert obs.shape == env.obs_shape

    def test_action_shape(self, env):
        assert hasattr(env, "action_shape")
        assert isinstance(env.action_shape, tuple)
        if env.action_shape:
            for val in env.action_shape:
                assert isinstance(val, int)

    def test_unwrapped(self, env):
        assert isinstance(env.unwrapped, PlanEnv)

    @pytest.mark.skipif(os.getenv("SKIP_RENDER", False), reason="No display in CI.")
    @pytest.mark.parametrize("return_image", [True, False])
    def test_return_image(self, env, return_image):
        assert isinstance(env.return_image, bool)
        if isinstance(env, VectorizedEnv):
            env.plan_env._return_image = return_image
        else:
            env._return_image = return_image
        _ = env.reset()
        *_, info = env.step(env.sample_action())
        if env.return_image:
            assert "rgb" in info

    # Test public API functions
    # ---------------------------------------------------------------------------------------------
    def test_sample_action(self, env):
        action = env.sample_action()
        if env.action_shape:
            assert action.shape == env.action_shape

    def test_get_state(self, env):
        state_reset, obs = env.reset()
        state = env.get_state()
        state_is_array = isinstance(state, numpy.ndarray)
        assert state_is_array if env.STATE_IS_ARRAY else not state_is_array
        if state_is_array and not env.SINGLETON:
            assert (state == state_reset).all(), f"original: {state} env: {env.get_state()}"

    def test_set_state(self, env):
        env.reset()
        state = env.get_state()
        env.step(env.sample_action())
        env.set_state(state)
        if env.STATE_IS_ARRAY:
            env_state = env.get_state()
            assert state.shape == env_state.shape
            if state.dtype != object and not env.SINGLETON:
                assert (state == env_state).all(), (state, env.get_state())

    def test_reset(self, env):
        _ = env.reset(return_state=False)
        state, obs = env.reset(return_state=True)
        state_is_array = isinstance(state, numpy.ndarray)
        obs_is_array = isinstance(obs, numpy.ndarray)
        assert state_is_array if env.STATE_IS_ARRAY else not state_is_array
        assert obs_is_array if env.OBS_IS_ARRAY else not obs_is_array

    @pytest.mark.parametrize("state", [None, True])
    @pytest.mark.parametrize("return_state", [None, True, False])
    def test_step(self, env, state, return_state, dt=1):
        _state, *_ = env.reset(return_state=True)
        if state is not None:
            state = _state
        action = env.sample_action()

        data = env.step(action, dt=dt, state=state, return_state=return_state)
        *new_state, obs, reward, terminal, info = data
        assert isinstance(data, tuple)
        # Test return state works correctly
        should_return_state = (return_state is None and state is not None) or return_state
        if should_return_state:
            assert len(new_state) == 1
            new_state = new_state[0]
            state_is_array = isinstance(new_state, numpy.ndarray)
            assert state_is_array if env.STATE_IS_ARRAY else not state_is_array
            if state_is_array:
                assert _state.shape == new_state.shape
            if not env.SINGLETON and env.STATE_IS_ARRAY:
                curr_state = env.get_state()
                assert (new_state == curr_state).all(), (
                    f"original: {new_state[new_state!= curr_state]} "
                    f"env: {curr_state[new_state!= curr_state]}"
                )
        else:
            assert len(new_state) == 0
        step_tuple_test(env, obs, reward, terminal, info, dt=dt)

    @pytest.mark.parametrize("states", [None, True, "None_list"])
    @pytest.mark.parametrize("return_state", [None, True, False])
    def test_step_batch(self, env, states, return_state, batch_size):
        dt = 1
        state, _ = env.reset()
        if states == "None_list":
            states = [None] * batch_size
        elif states:
            states = [copy.deepcopy(state) for _ in range(batch_size)]

        actions = [env.sample_action() for _ in range(batch_size)]

        data = env.step_batch(actions, dt=dt, states=states, return_state=return_state)
        *new_states, observs, rewards, terminals, infos = data
        assert isinstance(data, tuple)
        # Test return state works correctly
        default_returns_state = (
            return_state is None and isinstance(states, list) and states[0] is not None
        )
        should_return_state = return_state or default_returns_state
        if should_return_state:
            assert len(new_states) == 1
            new_states = new_states[0]
            # Todo: update check when returning batch arrays is available
            assert isinstance(new_states, list)
            state_is_array = isinstance(new_states[0], numpy.ndarray)
            assert state_is_array if env.STATE_IS_ARRAY else not state_is_array
            if env.STATE_IS_ARRAY:
                assert state.shape == new_states[0].shape
        else:
            assert len(new_states) == 0, (len(new_states), should_return_state, return_state)

        step_batch_tuple_test(
            env=env,
            batch_size=batch_size,
            observs=observs,
            rewards=rewards,
            terminals=terminals,
            infos=infos,
            dt=dt,
        )

    def test_step_dt_values(self, env, dt=3, return_state=None):
        state = None
        _state, *_ = env.reset(return_state=True)
        action = env.sample_action()

        data = env.step(action, dt=dt, state=state, return_state=return_state)
        *new_state, obs, reward, terminal, info = data
        assert isinstance(data, tuple)
        assert len(new_state) == 0
        step_tuple_test(env, obs, reward, terminal, info, dt=dt)

    @pytest.mark.parametrize("dt", [3, "array"])
    def test_step_batch_dt_values(self, env, dt, batch_size, states=None, return_state=None):
        dt = dt if dt != "array" else numpy.random.randint(1, 4, batch_size).astype(int)
        state, _ = env.reset()
        actions = [env.sample_action() for _ in range(batch_size)]

        data = env.step_batch(actions, dt=dt, states=states, return_state=return_state)
        *new_states, observs, rewards, terminals, infos = data
        assert isinstance(data, tuple)
        assert len(new_states) == 0, (len(new_states), return_state)

        step_batch_tuple_test(
            env=env,
            batch_size=batch_size,
            observs=observs,
            rewards=rewards,
            terminals=terminals,
            infos=infos,
            dt=dt,
        )

    @pytest.mark.filterwarnings("ignore::pytest.PytestUnraisableExceptionWarning")
    @pytest.mark.parametrize("delay_setup", [False, True])
    def test_clone_and_close(self, env, delay_setup):
        if not env.SINGLETON:
            env.delay_setup = delay_setup
            clone = env.clone()
            if clone.delay_setup:
                clone.reset()
            del clone

            clone = env.clone()
            if clone.delay_setup:
                clone.setup()
            clone.close()

    @pytest.mark.skipif(os.getenv("SKIP_RENDER", False), reason="No display in CI.")
    def test_get_image(self, env):
        img = env.get_image()
        if img is not None:
            assert isinstance(img, numpy.ndarray)
            assert len(img.shape) == 2 or len(img.shape) == 3


class TestPlangymEnv:
    CLASS_ATTRIBUTES = ("AVAILABLE_OBS_TYPES", "DEFAULT_OBS_TYPE")
    PROPERTIES = (
        "gym_env",
        "obs_shape",
        "obs_type",
        "observation_space",
        "action_shape",
        "action_space",
        "reward_range",
        "metadata",
        "render_mode",
        "remove_time_limit",
        "name",
        "frameskip",
        "autoreset",
        "delay_setup",
        "return_image",
    )

    def test_class_attributes(self, env):
        for name in self.CLASS_ATTRIBUTES:
            assert hasattr(env.__class__, name), f"Env {env.name} does not have attribute {name}"
            isinstance(getattr(env.__class__, name), bool)

    def test_has_properties(self, env):
        for name in self.PROPERTIES:
            assert hasattr(env, name), f"Env {env.name} does not have property {name}"

    def test_obs_type(self, env):
        assert isinstance(env.obs_type, str)
        assert env.obs_type in env.AVAILABLE_OBS_TYPES
        assert env.DEFAULT_OBS_TYPE in env.AVAILABLE_OBS_TYPES, (
            str(env.DEFAULT_OBS_TYPE),
            env.AVAILABLE_OBS_TYPES,
        )

    def test_obvervation_space(self, env):
        assert hasattr(env, "observation_space")
        assert isinstance(env.observation_space, gym.Space), (
            env.observation_space,
            env.DEFAULT_OBS_TYPE,
        )
        assert env.observation_space.shape == env.obs_shape
        if env.observation_space.shape:
            assert env.observation_space.shape == env.reset(return_state=False).shape

    def test_action_space(self, env):
        assert hasattr(env, "action_space")
        assert isinstance(env.action_space, gym.Space)
        assert env.action_space.shape == env.action_shape
        if env.action_space.shape:
            assert env.action_space.shape == env.sample_action().shape

    @pytest.mark.filterwarnings("ignore::pytest.PytestUnraisableExceptionWarning")
    def test_gym_env(self, env):
        assert hasattr(env.gym_env, "reset")
        assert hasattr(env.gym_env, "step")
        if not isinstance(env, VectorizedEnv) and not env.SINGLETON:
            env.close()
            env.gym_env

    def test_reward_range(self, env):
        env.reward_range

    @pytest.mark.parametrize("delay_setup", [True, False])
    def test_delay_setup(self, env, delay_setup):
        if env.SINGLETON or isinstance(env, VectorizedEnv):
            return
        new_env = env.clone(delay_setup=delay_setup)
        assert new_env._gym_env is None if delay_setup else new_env._gym_env is not None
        assert env.gym_env is not None

    def test_has_metadata(self, env):
        assert hasattr(env, "metadata")

    def test_render_mode(self, env):
        assert hasattr(env, "render_mode")
        if env.render_mode is not None:
            assert isinstance(env.render_mode, str)
        assert env.render_mode in env.AVAILABLE_RENDER_MODES

    def test_remove_time_limit(self, env):
        assert isinstance(env.remove_time_limit, bool)
        if env.remove_time_limit and not env._wrappers:
            assert "TimeLimit" not in str(env.gym_env), env.gym_env

    def test_seed(self, env):
        env.seed()
        env.seed(1)

    def test_terminal(self, env):
        if env.autoreset:
            env.reset()
            if hasattr(env, "render_mode") and env.render_mode in {"human", "rgb_array"}:
                return
            env.step_with_dt(env.sample_action(), dt=1000)

    @pytest.mark.skipif(os.getenv("SKIP_RENDER", False), reason="No display in CI.")
    def test_render(self, env, display):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            env.render()

    def test_wrap_environment(self, env):
        if isinstance(env, VectorizedEnv):
            return
        from gym.wrappers.transform_reward import TransformReward

        wrappers = [(TransformReward, {"f": lambda x: x})]
        env.apply_wrappers(wrappers)
        assert isinstance(env.gym_env, TransformReward)
        env._gym_env = env.gym_env.env

        wrappers = [(TransformReward, [lambda x: x])]
        env.apply_wrappers(wrappers)
        assert isinstance(env.gym_env, TransformReward)
        env._gym_env = env.gym_env.env

        wrappers = [(TransformReward, lambda x: x)]
        env.apply_wrappers(wrappers)
        assert isinstance(env.gym_env, TransformReward)
        env._gym_env = env.gym_env.env


class TestVideogameEnv:
    def test_ram(self, env):
        assert hasattr(env, "get_ram")
        assert isinstance(env.get_ram(), numpy.ndarray)



import os
import warnings

import numpy
import pytest
import ray

from plangym.control.classic_control import ClassicControl
from plangym.vectorization.ray import RayEnv, RemoteEnv
from plangym.videogames.atari import AtariEnv


pytest.importorskip("ray")
if os.getenv("DISABLE_RAY", False) and str(os.getenv("DISABLE_RAY", "False")).lower() != "false":
    pytest.skip("Ray not installed or disabled", allow_module_level=True)
from plangym.api_tests import batch_size, display, TestPlanEnv, TestPlangymEnv  # noqa: F401


def ray_cartpole():
    return RayEnv(env_class=ClassicControl, name="CartPole-v0", n_workers=2)


def ray_retro():
    from plangym.videogames.retro import RetroEnv

    return RayEnv(env_class=RetroEnv, name="Airstriker-Genesis", n_workers=2)


def ray_dm_control():
    from plangym.control.dm_control import DMControlEnv

    return RayEnv(env_class=DMControlEnv, name="walker-walk", n_workers=2)


environments = [(ray_cartpole, True), (ray_retro, False), (ray_dm_control, True)]


@pytest.fixture(params=environments, scope="module")
def env(request) -> AtariEnv:
    env_call, local = request.param
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        ray.init(ignore_reinit_error=True, local_mode=local)
    yield env_call()
    ray.shutdown()


def test_remote_actor():
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        ray.init(ignore_reinit_error=True, local_mode=True)

    def create_cartpole():
        return ClassicControl(name="CartPole-v0")

    env = RemoteEnv.remote(create_cartpole)
    ray.get(env.setup.remote())
    ray.get(env.reset.remote())
    ray.get(env.step.remote(0))
    state = ray.get(env.get_state.remote())
    assert isinstance(state, numpy.ndarray)



"""Environment for playing Mario bros using gym-super-mario-bros."""
from typing import Any, Dict, Optional

import gym
import numpy

from plangym.videogames.env import VideogameEnv


# actions for the simple run right environment
RIGHT_ONLY = [
    ["NOOP"],
    ["right"],
    ["right", "A"],
    ["right", "B"],
    ["right", "A", "B"],
]


# actions for very simple movement
SIMPLE_MOVEMENT = [
    ["NOOP"],
    ["right"],
    ["right", "A"],
    ["right", "B"],
    ["right", "A", "B"],
    ["A"],
    ["left"],
]


# actions for more complex movement
COMPLEX_MOVEMENT = [
    ["NOOP"],
    ["right"],
    ["right", "A"],
    ["right", "B"],
    ["right", "A", "B"],
    ["A"],
    ["left"],
    ["left", "A"],
    ["left", "B"],
    ["left", "A", "B"],
    ["down"],
    ["up"],
]


class NesEnv(VideogameEnv):
    """Environment for working with the NES-py emulator."""

    @property
    def nes_env(self) -> "NESEnv":  # noqa: F821
        """Access the underlying NESEnv."""
        return self.gym_env.unwrapped

    def get_image(self) -> numpy.ndarray:
        """
        Return a numpy array containing the rendered view of the environment.

        Square matrices are interpreted as a greyscale image. Three-dimensional arrays
        are interpreted as RGB images with channels (Height, Width, RGB)
        """
        return self.gym_env.screen.copy()

    def get_ram(self) -> numpy.ndarray:
        """Return a copy of the emulator environment."""
        return self.nes_env.ram.copy()

    def get_state(self, state: Optional[numpy.ndarray] = None) -> numpy.ndarray:
        """
        Recover the internal state of the simulation.

        A state must completely describe the Environment at a given moment.
        """
        return self.gym_env.get_state(state)

    def set_state(self, state: numpy.ndarray) -> None:
        """
        Set the internal state of the simulation.

        Args:
            state: Target state to be set in the environment.

        Returns:
            None

        """
        self.gym_env.set_state(state)

    def close(self) -> None:
        """Close the underlying :class:`gym.Env`."""
        if self.nes_env._env is None:
            return
        try:
            super(NesEnv, self).close()
        except ValueError:  # pragma: no cover
            pass

    def __del__(self):
        """Tear down the environment."""
        try:
            self.close()
        except ValueError:  # pragma: no cover
            pass


class MarioEnv(NesEnv):
    """Interface for using gym-super-mario-bros in plangym."""

    AVAILABLE_OBS_TYPES = {"coords", "rgb", "grayscale", "ram"}
    MOVEMENTS = {
        "complex": COMPLEX_MOVEMENT,
        "simple": SIMPLE_MOVEMENT,
        "right": RIGHT_ONLY,
    }

    def __init__(
        self,
        name: str,
        movement_type: str = "simple",
        original_reward: bool = False,
        **kwargs,
    ):
        """
        Initialize a MarioEnv.

        Args:
            name: Name of the environment.
            movement_type: One of {complex|simple|right}
            original_reward: If False return a custom reward based on mario position and level.
            **kwargs: passed to super().__init__.
        """
        self._movement_type = movement_type
        self._original_reward = original_reward
        super(MarioEnv, self).__init__(name=name, **kwargs)

    def get_state(self, state: Optional[numpy.ndarray] = None) -> numpy.ndarray:
        """
        Recover the internal state of the simulation.

        A state must completely describe the Environment at a given moment.
        """
        state = numpy.empty(250288, dtype=numpy.byte) if state is None else state
        state[-2:] = 0  # Some states use the last two bytes. Set to zero by default.
        return super(MarioEnv, self).get_state(state)

    def init_gym_env(self) -> gym.Env:
        """Initialize the :class:`NESEnv`` instance that the current class is wrapping."""
        import gym_super_mario_bros
        from gym_super_mario_bros.actions import COMPLEX_MOVEMENT  # , SIMPLE_MOVEMENT
        from nes_py.wrappers import JoypadSpace

        env = gym_super_mario_bros.make(self.name)
        gym_env = JoypadSpace(env.unwrapped, COMPLEX_MOVEMENT)
        gym_env.reset()
        return gym_env

    def _update_info(self, info: Dict[str, Any]) -> Dict[str, Any]:
        info["player_state"] = self.nes_env._player_state
        info["area"] = self.nes_env._area
        info["left_x_position"] = self.nes_env._left_x_position
        info["is_stage_over"] = self.nes_env._is_stage_over
        info["is_dying"] = self.nes_env._is_dying
        info["is_dead"] = self.nes_env._is_dead
        info["y_pixel"] = self.nes_env._y_pixel
        info["y_viewport"] = self.nes_env._y_viewport
        info["x_position_last"] = self.nes_env._x_position_last
        info["in_pipe"] = (info["player_state"] == 0x02) or (info["player_state"] == 0x03)
        return info

    def _get_info(
        self,
    ):
        info = {
            "x_pos": 0,
            "y_pos": 0,
            "world": 0,
            "stage": 0,
            "life": 0,
            "coins": 0,
            "flag_get": False,
            "in_pipe": False,
        }
        return self._update_info(info)

    def get_coords_obs(
        self,
        obs: numpy.ndarray,
        info: Dict[str, Any] = None,
        **kwargs,
    ) -> numpy.ndarray:
        """Return the information contained in info as an observation if obs_type == "info"."""
        if self.obs_type == "coords":
            info = info or self._get_info()
            obs = numpy.array(
                [
                    info.get("x_pos", 0),
                    info.get("y_pos", 0),
                    info.get("world" * 10, 0),
                    info.get("stage", 0),
                    info.get("life", 0),
                    int(info.get("flag_get", 0)),
                    info.get("coins", 0),
                ],
            )
        return obs

    def process_reward(self, reward, info, **kwargs) -> float:
        """Return a custom reward based on the x, y coordinates and level mario is in."""
        if not self._original_reward:
            reward = (
                (info.get("world", 0) * 25000)
                + (info.get("stage", 0) * 5000)
                + info.get("x_pos", 0)
                + 10 * int(bool(info.get("in_pipe", 0)))
                + 100 * int(bool(info.get("flag_get", 0)))
                # + (abs(info["x_pos"] - info["x_position_last"]))
            )
        return reward

    def process_terminal(self, terminal, info, **kwargs) -> bool:
        """Return True if terminal or mario is dying."""
        return terminal or info.get("is_dying", False) or info.get("is_dead", False)

    def process_info(self, info, **kwargs) -> Dict[str, Any]:
        """Add additional data to the info dictionary."""
        return self._update_info(info)
